import copy
import math
import random
from datetime import datetime, timedelta, timezone

import numpy as np
import pandas as pd

from solie.definition.simluation_error import SimulationError
from solie.recipe import decide


def do(dataset):
    # leverage is treated as 1
    # because those are going to be applied at the presentation phase

    # ■■■■■ get data ■■■■■

    progress_list = dataset["progress_list"]
    target_progress = dataset["target_progress"]
    target_symbols = dataset["target_symbols"]
    calculation_index = dataset["calculation_index"]
    chunk_candle_data = dataset["chunk_candle_data"]
    chunk_indicators = dataset["chunk_indicators"]
    chunk_asset_record = dataset["chunk_asset_record"]
    chunk_unrealized_changes = dataset["chunk_unrealized_changes"]
    chunk_scribbles = dataset["chunk_scribbles"]
    chunk_account_state = dataset["chunk_account_state"]
    chunk_virtual_state = dataset["chunk_virtual_state"]
    decision_script = dataset["decision_script"]

    # ■■■■■ basic values ■■■■■

    decision_lag = 3000  # milliseconds

    # ■■■■■ return blank data if there's nothing to calculate ■■■■■

    if len(calculation_index) == 0:
        dataset = {
            "chunk_asset_record": chunk_asset_record,
            "chunk_unrealized_changes": chunk_unrealized_changes,
            "chunk_scribbles": chunk_scribbles,
            "chunk_account_state": chunk_account_state,
            "chunk_virtual_state": chunk_virtual_state,
        }

        return dataset

    # ■■■■■ convert to numpy objects for fast calculation ■■■■■

    calculation_index_ar = calculation_index.to_numpy()  # inside are datetime objects
    candle_data_ar = chunk_candle_data.to_records()
    indicators_ar = chunk_indicators.to_records()
    asset_record_ar = chunk_asset_record.to_records()
    chunk_unrealized_changes_ar = chunk_unrealized_changes.to_frame().to_records()

    # ■■■■■ actual loop calculation ■■■■■

    calculation_index_length = len(calculation_index_ar)
    decision_script = compile(decision_script, "<string>", "exec")
    first_calculation_moment = calculation_index_ar[0]

    for cycle in range(calculation_index_length):
        before_moment = calculation_index_ar[cycle]
        current_moment = before_moment + timedelta(seconds=10)

        for symbol in target_symbols:
            # ■■■■■ basic variables ■■■■■

            open_price = candle_data_ar[cycle][str((symbol, "Open"))]
            close_price = candle_data_ar[cycle][str((symbol, "Close"))]
            if math.isnan(open_price) or math.isnan(close_price):
                continue

            would_trade_happen = False
            did_found_new_trade = False
            amount_shift = 0
            fill_price = 0
            role = ""

            # ■■■■■ check if any order would be filled ■■■■■

            price_speed = (close_price - open_price) / 10
            is_margin_negative = False
            is_margin_nan = False

            # special placements
            if "cancel_all" in chunk_virtual_state["placements"][symbol]:
                cancel_placement_names = []
                for placement_name in chunk_virtual_state["placements"][symbol].keys():
                    if any(s in placement_name for s in ("later", "book")):
                        cancel_placement_names.append(placement_name)
                for cancel_placement_name in cancel_placement_names:
                    chunk_virtual_state["placements"][symbol].pop(cancel_placement_name)
                chunk_virtual_state["placements"][symbol].pop("cancel_all")

            # instant placements
            if "now_close" in chunk_virtual_state["placements"][symbol]:
                would_trade_happen = True
                command = chunk_virtual_state["placements"][symbol]["now_close"]
                role = "taker"
                fill_price = open_price + price_speed * (decision_lag / 1000)
                amount_shift = -chunk_virtual_state["locations"][symbol]["amount"]
                chunk_virtual_state["placements"][symbol].pop("now_close")

            if "now_buy" in chunk_virtual_state["placements"][symbol]:
                would_trade_happen = True
                command = chunk_virtual_state["placements"][symbol]["now_buy"]
                role = "taker"
                fill_price = open_price + price_speed * (decision_lag / 1000)
                fill_margin = command["margin"]
                if fill_margin < 0:
                    is_margin_negative = True
                if math.isnan(fill_margin):
                    is_margin_nan = True
                amount_shift = fill_margin / fill_price
                chunk_virtual_state["placements"][symbol].pop("now_buy")

            if "now_sell" in chunk_virtual_state["placements"][symbol]:
                would_trade_happen = True
                command = chunk_virtual_state["placements"][symbol]["now_sell"]
                role = "taker"
                fill_price = open_price + price_speed * (decision_lag / 1000)
                fill_margin = command["margin"]
                if fill_margin < 0:
                    is_margin_negative = True
                if math.isnan(fill_margin):
                    is_margin_nan = True
                amount_shift = -fill_margin / fill_price
                chunk_virtual_state["placements"][symbol].pop("now_sell")

            # conditional placements
            if "later_up_close" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["later_up_close"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "taker"
                    fill_price = boundary
                    amount_shift = -chunk_virtual_state["locations"][symbol]["amount"]
                    chunk_virtual_state["placements"][symbol].pop("later_up_close")

            if "later_down_close" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["later_down_close"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "taker"
                    fill_price = boundary
                    amount_shift = -chunk_virtual_state["locations"][symbol]["amount"]
                    chunk_virtual_state["placements"][symbol].pop("later_down_close")

            if "later_up_buy" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["later_up_buy"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "taker"
                    fill_price = boundary
                    fill_margin = command["margin"]
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = fill_margin / fill_price
                    chunk_virtual_state["placements"][symbol].pop("later_up_buy")

            if "later_down_buy" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["later_down_buy"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "taker"
                    fill_price = boundary
                    fill_margin = command["margin"]
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = fill_margin / fill_price
                    chunk_virtual_state["placements"][symbol].pop("later_down_buy")

            if "later_up_sell" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["later_up_sell"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "taker"
                    fill_price = boundary
                    fill_margin = command["margin"]
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = -fill_margin / fill_price
                    chunk_virtual_state["placements"][symbol].pop("later_up_sell")

            if "later_down_sell" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["later_down_sell"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "taker"
                    fill_price = boundary
                    fill_margin = command["margin"]
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = -fill_margin / fill_price
                    chunk_virtual_state["placements"][symbol].pop("later_down_sell")

            if "book_buy" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["book_buy"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "maker"
                    fill_price = boundary
                    fill_margin = command["margin"]
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = fill_margin / fill_price
                    chunk_virtual_state["placements"][symbol].pop("book_buy")

            if "book_sell" in chunk_virtual_state["placements"][symbol]:
                command = chunk_virtual_state["placements"][symbol]["book_sell"]
                boundary = command["boundary"]

                wobble_high = candle_data_ar[cycle][str((symbol, "High"))]
                wobble_low = candle_data_ar[cycle][str((symbol, "Low"))]
                did_cross = wobble_low < boundary < wobble_high

                if did_cross:
                    would_trade_happen = True
                    role = "maker"
                    fill_price = boundary
                    fill_margin = command["margin"]
                    if fill_margin < 0:
                        is_margin_negative = True
                    if math.isnan(fill_margin):
                        is_margin_nan = True
                    amount_shift = -fill_margin / fill_price
                    chunk_virtual_state["placements"][symbol].pop("book_sell")

            # check if situation is okay
            if is_margin_negative:
                text = ""
                text += "Got an order with a negative margin"
                text += f" while calculating {symbol} market at {current_moment}"
                raise SimulationError(text)

            elif is_margin_nan:
                text = ""
                text += "Got an order with a non-numeric margin"
                text += f" while calculating {symbol} market at {current_moment}"
                raise SimulationError(text)

            # ■■■■■ mimic the real world phenomenon ■■■■■

            if would_trade_happen:
                symbol_location = chunk_virtual_state["locations"][symbol]
                before_entry_price = symbol_location["entry_price"]
                before_amount = symbol_location["amount"]

                symbol_location["amount"] += amount_shift
                current_amount = symbol_location["amount"]

                # case when the position is created from 0
                if before_amount == 0 and current_amount != 0:
                    symbol_location["entry_price"] = fill_price
                    invested_margin = abs(current_amount) * fill_price
                    chunk_virtual_state["available_balance"] -= invested_margin
                # case when the position is closed from something
                elif before_amount != 0 and current_amount == 0:
                    symbol_location["entry_price"] = 0
                    price_difference = fill_price - before_entry_price
                    realized_profit = price_difference * before_amount
                    returned_margin = abs(before_amount) * before_entry_price
                    chunk_virtual_state["available_balance"] += returned_margin
                    chunk_virtual_state["available_balance"] += realized_profit
                # case when the position direction is flipped
                elif before_amount * current_amount < 0:
                    symbol_location["entry_price"] = fill_price
                    price_difference = fill_price - before_entry_price
                    realized_profit = price_difference * before_amount
                    returned_margin = abs(before_amount) * before_entry_price
                    invested_margin = abs(current_amount) * fill_price
                    chunk_virtual_state["available_balance"] += returned_margin
                    chunk_virtual_state["available_balance"] -= invested_margin
                    chunk_virtual_state["available_balance"] += realized_profit
                # case when the position size is increased one the same direction
                elif abs(current_amount) > abs(before_amount):
                    before_numerator = before_entry_price * before_amount
                    new_numerator = fill_price * amount_shift
                    current_numerator = before_numerator + new_numerator
                    new_entry_price = current_numerator / current_amount
                    symbol_location["entry_price"] = new_entry_price
                    realized_profit = 0
                    invested_margin = abs(amount_shift) * fill_price
                    chunk_virtual_state["available_balance"] -= invested_margin
                    chunk_virtual_state["available_balance"] += realized_profit
                # case when the position size is decreased one the same direction
                else:
                    symbol_location["entry_price"] = before_entry_price
                    price_difference = fill_price - before_entry_price
                    realized_profit = price_difference * (-amount_shift)
                    returned_margin = abs(amount_shift) * before_entry_price
                    chunk_virtual_state["available_balance"] += returned_margin
                    chunk_virtual_state["available_balance"] += realized_profit

                did_found_new_trade = True

                if chunk_virtual_state["available_balance"] < 0:
                    text = ""
                    text += "Available balance went below zero"
                    text += f" while calculating {symbol} market"
                    text += f" at {current_moment}"
                    raise SimulationError(text)

            # ■■■■■ update the account state (symbol dependent) ■■■■■

            # locations
            current_entry_price = chunk_virtual_state["locations"][symbol][
                "entry_price"
            ]
            current_entry_price = float(current_entry_price)
            current_amount = chunk_virtual_state["locations"][symbol]["amount"]
            current_margin = abs(current_amount) * current_entry_price
            current_margin = float(current_margin)
            chunk_account_state["positions"][symbol][
                "entry_price"
            ] = current_entry_price
            chunk_account_state["positions"][symbol]["margin"] = current_margin
            if chunk_virtual_state["locations"][symbol]["amount"] > 0:
                chunk_account_state["positions"][symbol]["direction"] = "long"
            if chunk_virtual_state["locations"][symbol]["amount"] < 0:
                chunk_account_state["positions"][symbol]["direction"] = "short"
            if chunk_virtual_state["locations"][symbol]["amount"] == 0:
                chunk_account_state["positions"][symbol]["direction"] = "none"

            # placements
            symbol_placements = chunk_virtual_state["placements"][symbol]
            symbol_open_orders = {}
            for command_name, placement in symbol_placements.items():
                order_id = placement["order_id"]
                boundary = float(placement["boundary"])
                if "margin" in placement.keys():
                    left_margin = float(placement["margin"])
                else:
                    left_margin = None
                symbol_open_orders[order_id] = {
                    "command_name": command_name,
                    "boundary": boundary,
                    "left_margin": left_margin,
                }
            chunk_account_state["open_orders"][symbol] = symbol_open_orders

            # ■■■■■ record (symbol dependent) ■■■■■

            if did_found_new_trade:
                fill_time = before_moment + timedelta(milliseconds=decision_lag)
                fill_time = np.datetime64(fill_time)
                while fill_time in asset_record_ar["index"]:
                    fill_time += np.timedelta64(1, "ms")

                wallet_balance = chunk_virtual_state["available_balance"]
                for symbol_key, location in chunk_virtual_state["locations"].items():
                    if location["amount"] == 0:
                        continue
                    column_key = str((symbol_key, "Close"))
                    symbol_price = candle_data_ar[cycle][column_key]
                    if math.isnan(symbol_price):
                        continue
                    current_margin = abs(location["amount"]) * location["entry_price"]
                    wallet_balance += current_margin

                margin_ratio = abs(amount_shift) * open_price / wallet_balance

                order_id = random.randint(10**18, 10**19 - 1)

                if amount_shift > 0:
                    side = "buy"
                elif amount_shift < 0:
                    side = "sell"
                else:
                    raise ValueError("Amount of asset shift cannot be 0")

                if role == "":
                    raise ValueError("No trade role was specified")

                if fill_price == 0:
                    raise ValueError("The fill price cannot be zero")

                original_size = asset_record_ar.shape[0]
                asset_record_ar.resize(original_size + 1)
                asset_record_ar[-1]["index"] = fill_time
                asset_record_ar[-1]["Cause"] = "auto_trade"
                asset_record_ar[-1]["Symbol"] = symbol
                asset_record_ar[-1]["Side"] = side
                asset_record_ar[-1]["Fill Price"] = fill_price
                asset_record_ar[-1]["Role"] = role
                asset_record_ar[-1]["Margin Ratio"] = margin_ratio
                asset_record_ar[-1]["Order ID"] = order_id
                asset_record_ar[-1]["Result Asset"] = wallet_balance

                update_time = fill_time.astype(datetime).replace(tzinfo=timezone.utc)
                chunk_account_state["positions"][symbol]["update_time"] = update_time

        # ■■■■■ understand the situation ■■■■■

        wallet_balance = chunk_virtual_state["available_balance"]
        unrealized_profit = 0
        for symbol_key, location in chunk_virtual_state["locations"].items():
            if location["amount"] == 0:
                continue
            symbol_price = candle_data_ar[cycle][str((symbol_key, "Close"))]
            if math.isnan(symbol_price):
                continue
            current_margin = abs(location["amount"]) * location["entry_price"]
            wallet_balance += current_margin
            # assume that mark price doesn't wobble more than 5%
            key_open_price = candle_data_ar[cycle][str((symbol_key, "Open"))]
            key_close_price = candle_data_ar[cycle][str((symbol_key, "Close"))]
            if location["amount"] < 0:
                basic_price = max(key_open_price, key_close_price) * 1.05
                key_high_price = candle_data_ar[cycle][str((symbol_key, "High"))]
                extreme_price = min(basic_price, key_high_price)
            else:
                basic_price = min(key_open_price, key_close_price) * 0.95
                key_low_price = candle_data_ar[cycle][str((symbol_key, "Low"))]
                extreme_price = max(basic_price, key_low_price)
            price_difference = extreme_price - location["entry_price"]
            unrealized_profit += price_difference * location["amount"]
        unrealized_change = unrealized_profit / wallet_balance

        # ■■■■■ update the account state (symbol independent) ■■■■■

        chunk_account_state["observed_until"] = current_moment
        chunk_account_state["wallet_balance"] = float(wallet_balance)

        # ■■■■■ record (symbol independent) ■■■■■

        original_size = chunk_unrealized_changes_ar.shape[0]
        chunk_unrealized_changes_ar.resize(original_size + 1)
        chunk_unrealized_changes_ar[-1]["index"] = before_moment
        chunk_unrealized_changes_ar[-1]["0"] = unrealized_change

        # ■■■■■ make decision and place order ■■■■■

        current_candle_data = candle_data_ar[cycle]
        current_indicators = indicators_ar[cycle]
        decision, chunk_scribbles = decide.choose(
            target_symbols=target_symbols,
            current_moment=current_moment,
            current_candle_data=current_candle_data,
            current_indicators=current_indicators,
            account_state=copy.deepcopy(chunk_account_state),
            scribbles=chunk_scribbles,
            decision_script=decision_script,
        )

        for symbol_key, symbol_decision in decision.items():
            for each_decision in symbol_decision.values():
                each_decision["order_id"] = random.randint(10**18, 10**19 - 1)
            chunk_virtual_state["placements"][symbol_key].update(decision[symbol_key])

        # ■■■■■ report the progress in seconds ■■■■■

        progress_list[target_progress] = max(
            (current_moment - first_calculation_moment).total_seconds(), 0
        )

    # ■■■■■ convert back numpy objects to pandas objects ■■■■■

    chunk_asset_record = pd.DataFrame(asset_record_ar)
    chunk_asset_record = chunk_asset_record.set_index("index")
    chunk_asset_record.index.name = None
    chunk_asset_record.index = pd.to_datetime(chunk_asset_record.index, utc=True)

    chunk_unrealized_changes = pd.DataFrame(chunk_unrealized_changes_ar)
    chunk_unrealized_changes = chunk_unrealized_changes.set_index("index")
    chunk_unrealized_changes.index.name = None
    chunk_unrealized_changes.index = pd.to_datetime(
        chunk_unrealized_changes.index, utc=True
    )
    chunk_unrealized_changes = chunk_unrealized_changes["0"]

    # ■■■■■ return calculated data ■■■■■

    dataset = {
        "chunk_asset_record": chunk_asset_record,
        "chunk_unrealized_changes": chunk_unrealized_changes,
        "chunk_scribbles": chunk_scribbles,
        "chunk_account_state": chunk_account_state,
        "chunk_virtual_state": chunk_virtual_state,
    }

    return dataset



import hashlib
import hmac
from datetime import datetime, timezone
from urllib.parse import urlencode

import aiohttp

from solie.definition.api_request_error import ApiRequestError


class ApiRequester:
    used_rates = {}

    def __init__(self):
        self.keys = {
            "binance_api": "",
            "binance_secret": "",
        }

    def update_keys(self, keys):
        self.keys.update(keys)

    async def binance(
        self, http_method: str, path: str, payload: dict = {}, server="futures"
    ):
        query_string = urlencode(payload)
        # replace single quote to double quote
        query_string = query_string.replace("%27", "%22")

        signature = hmac.new(
            self.keys["binance_secret"].encode("utf-8"),
            query_string.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()
        headers = {"X-MBX-APIKEY": self.keys["binance_api"]}

        if server == "spot":
            url = "https://api.binance.com"
        elif server == "futures":
            url = "https://fapi.binance.com"
        else:
            raise ApiRequestError("This Binance server is supported")
        url += path
        url += "?" + query_string + "&signature=" + signature

        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=http_method, url=url, headers=headers
            ) as raw_response:
                response = await raw_response.json()

        # record api usage
        for header_key in raw_response.headers.keys():
            if "X-MBX" in header_key:
                write_value = raw_response.headers[header_key]
                current_time = datetime.now(timezone.utc)
                self.used_rates[header_key] = (write_value, current_time)

        # check if the response contains error message
        if "code" in response and response["code"] != 200:
            error_code = response["code"]
            error_message = response["msg"]
            text = "There was a problem with Binance API request"
            text += f" (Error {error_code}: {error_message})"
            raise ApiRequestError(text)

        return response

    async def coingecko(self, http_method: str, path: str, payload: dict = {}):
        query_string = urlencode(payload)
        # replace single quote to double quote
        query_string = query_string.replace("%27", "%22")

        url = "https://api.coingecko.com" + path + "?" + query_string

        async with aiohttp.ClientSession() as session:
            async with session.request(method=http_method, url=url) as raw_response:
                response = await raw_response.json()

        return response

    async def bytes(self, url: str):
        headers = {
            "User-agent": "Mozilla/5.0",
        }

        async with aiohttp.ClientSession() as session:
            async with session.request(
                method="GET", url=url, headers=headers
            ) as raw_response:
                response = await raw_response.read()

        status_code = raw_response.status
        if status_code != 200:
            text = f"There was a problem with bytes request (HTTP {status_code})"
            text += "\n"
            text += url
            raise ApiRequestError(text)

        return response



import asyncio
import logging
import os
import sys
import time
from datetime import datetime, timezone

import aiofiles

import solie
from solie import introduction
from solie.overlay.long_text_view import LongTextView


class LogHandler(logging.Handler):
    def __init__(self):
        super().__init__()

        self.executed_time = datetime.now(timezone.utc).replace(microsecond=0)

        log_format = "%(asctime)s.%(msecs)03d %(levelname)s"
        date_format = "%Y-%m-%d %H:%M:%S"
        log_formatter = logging.Formatter(log_format, datefmt=date_format)
        log_formatter.converter = time.gmtime
        self.setFormatter(log_formatter)

        os.makedirs(f"{introduction.PATH}/note/logs", exist_ok=True)
        now = self.executed_time
        self.filepath = (
            f"{introduction.PATH}/note/logs"
            + f"/{now.year:04}-{now.month:02}-{now.day:02}"
            + f".{now.hour:02}-{now.minute:02}-{now.second:02}"
            + f".{now.tzinfo}.txt"
        )

    def emit(self, log_record: logging.LogRecord):
        formatted = self.format(log_record)

        if log_record.exc_info is not None:
            # when this is from an exception
            lines = formatted.split("\n")
            summarization = lines[0]
            log_content = "\n".join(lines[1:])
            exception = log_record.exc_info[0]
            if exception is None:
                return
            exc_type = exception.__name__
            summarization += f" - {exc_type}"
        else:
            # when this is a normal log
            summarization = formatted
            log_content = log_record.getMessage()
            first_line_content = log_content.split("\n")[0].strip()
            summarization += f" - {first_line_content}"
            summarization = summarization[:80]

        if not solie.app_close_event.is_set():
            if solie.window.should_overlap_error:

                async def job(log_content=log_content):
                    formation = [
                        "There was an error",
                        LongTextView,
                        False,
                        [log_content],
                    ]
                    await solie.window.overlay(formation)

                asyncio.create_task(job())

            else:
                asyncio.create_task(
                    self.add_log_output(
                        summarization,
                        log_content,
                    ),
                )
        else:
            sys.stdout.write(log_content)

    async def add_log_output(self, *args, **kwargs):
        # get the data
        summarization = args[0]
        log_content = args[1]

        # add to log list
        solie.window.listWidget.addItem(summarization, log_content)

        # save to file
        async with aiofiles.open(self.filepath, "a", encoding="utf8") as file:
            await file.write(f"{summarization}\n")
            await file.write(f"{log_content}\n\n")



import asyncio
from typing import Callable, Coroutine

import aiohttp

import solie


class ApiStreamer:
    def __init__(self, url: str, when_received: Callable[..., Coroutine]):
        self._url = url
        self._when_received = when_received
        self.session = aiohttp.ClientSession()

        if url != "":
            asyncio.create_task(self._run_websocket())

    def __del__(self):
        asyncio.create_task(self._close_self())

    async def _run_websocket(self):
        async with self.session.ws_connect(self._url) as ws:
            async for received_raw in ws:
                if solie.app_close_event.is_set():
                    return
                if received_raw.type in (
                    aiohttp.WSMsgType.CLOSED,
                    aiohttp.WSMsgType.ERROR,
                ):
                    asyncio.create_task(self._run_websocket())
                    raise ConnectionError(f"Websocket reconnected:\n{self._url}")
                else:
                    received = received_raw.json()
                    asyncio.create_task(self._when_received(received=received))

    async def _close_self(self):
        await self.session.close()
        solie.logger.exception(f"Websocket closed:\n{self._url}")



import copy
import json
import os

import aiofiles

from solie import introduction

_app_settings = {
    "datapath": None,
}
_data_settings = {
    "asset_token": None,
    "target_symbols": None,
}


async def load():
    global _app_settings
    global _data_settings

    filepath = f"{introduction.PATH}/note/app_settings.json"
    if os.path.isfile(filepath):
        async with aiofiles.open(filepath, "r", encoding="utf8") as file:
            content = await file.read()
            _app_settings = json.loads(content)

    datapath = _app_settings["datapath"]
    filepath = f"{datapath}/data_settings.json"
    if os.path.isfile(filepath):
        async with aiofiles.open(filepath, "r", encoding="utf8") as file:
            content = await file.read()
            _data_settings = json.loads(content)


def get_app_settings() -> dict:
    return copy.deepcopy(_app_settings)


async def apply_app_settings(payload):
    global _app_settings
    payload = copy.deepcopy(payload)
    _app_settings = {**_app_settings, **payload}
    filepath = f"{introduction.PATH}/note/app_settings.json"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    async with aiofiles.open(filepath, "w", encoding="utf8") as file:
        content = json.dumps(_app_settings, indent=4)
        await file.write(content)


def get_data_settings() -> dict:
    return copy.deepcopy(_data_settings)


async def apply_data_settings(payload):
    global _data_settings
    payload = copy.deepcopy(payload)
    _data_settings = {**_data_settings, **payload}
    datapath = _app_settings["datapath"]
    if datapath is not None:
        filepath = f"{datapath}/data_settings.json"
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        async with aiofiles.open(filepath, "w", encoding="utf8") as file:
            content = json.dumps(_data_settings, indent=4)
            await file.write(content)



from PySide6 import QtWidgets

from solie.recipe import outsource


class DownloadFillOption(QtWidgets.QWidget):
    def __init__(self, done_event, payload):
        # ■■■■■ the basic ■■■■■

        super().__init__()

        # ■■■■■ prepare variables ■■■■■

        answer_container = payload[0]
        fill_options = (
            "From 2020 to last year",
            "From 2020 to last year (With small daily files)",
            "From first month of this year to last month",
            "From first month of this year to last month (With small daily files)",
            "This month",
            "Yesterday and the day before yesterday",
        )

        # ■■■■■ full layout ■■■■■

        full_layout = QtWidgets.QHBoxLayout(self)
        cards_layout = QtWidgets.QVBoxLayout()
        full_layout.addLayout(cards_layout)

        # ■■■■■ spacing ■■■■■

        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)

        # ■■■■■ a card ■■■■■

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QVBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # explanation
        explain_label = QtWidgets.QLabel(
            "Solie will fill the candle data with historical data provided by"
            " Binance. The more you fill, the longer it takes. Amount of a few days"
            " only takes few minutes while amount of a few years can take hours."
        )
        explain_label.setWordWrap(True)
        card_layout.addWidget(explain_label)

        # ■■■■■ a card ■■■■■

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QVBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # option buttons
        for turn, text in enumerate(fill_options):

            async def job(turn=turn, *args, **kwargs):
                answer_container["filling_type"] = turn
                done_event.set()

            option_button = QtWidgets.QPushButton(text, card)
            outsource.do(option_button.clicked, job)
            card_layout.addWidget(option_button)

        # ■■■■■ spacing ■■■■■

        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)



def keys(dictionary_object, target_value):
    keys = []
    for item_key, item_value in dictionary_object.items():
        if item_value == target_value:
            keys.append(item_key)
    return keys


def indexes(iterable, target_value):
    indexes = []
    for item_index, item_value in enumerate(iterable):
        if item_value == target_value:
            indexes.append(item_index)
    return indexes



import re

from PySide6 import QtCore, QtGui, QtWidgets

import solie
from solie import introduction
from solie.recipe import compare_versions, outsource
from solie.widget.horizontal_divider import HorizontalDivider


class StrategyBasicInput(QtWidgets.QWidget):
    def __init__(self, done_event, payload):
        # ■■■■■ the basic ■■■■■

        super().__init__()

        # ■■■■■ prepare things ■■■■■

        strategy = payload

        # ■■■■■ full layout ■■■■■

        full_layout = QtWidgets.QHBoxLayout(self)
        cards_layout = QtWidgets.QVBoxLayout()
        full_layout.addLayout(cards_layout)

        # ■■■■■ spacing ■■■■■

        # spacing
        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)

        # ■■■■■ a card ■■■■■

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QVBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # explanation
        detail_text = QtWidgets.QLabel()
        detail_text.setText("About")
        detail_text.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        detail_text.setWordWrap(True)
        card_layout.addWidget(detail_text)

        # spacing
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(3)
        spacing_text.setFont(spacing_text_font)
        card_layout.addWidget(spacing_text)

        # divider
        divider = HorizontalDivider(self)
        card_layout.addWidget(divider)

        # spacing
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(3)
        spacing_text.setFont(spacing_text_font)
        card_layout.addWidget(spacing_text)

        # input
        this_layout = QtWidgets.QFormLayout()
        card_layout.addLayout(this_layout)
        code_name_input = QtWidgets.QLineEdit()
        code_name_input.setText(strategy["code_name"])
        this_layout.addRow("Code name", code_name_input)
        readable_name_input = QtWidgets.QLineEdit()
        readable_name_input.setText(strategy["readable_name"])
        this_layout.addRow("Readable name", readable_name_input)
        version_input = QtWidgets.QLineEdit()
        version_input.setText(strategy["version"])
        this_layout.addRow("Version", version_input)
        description_input = QtWidgets.QTextEdit()
        description_input.setPlainText(strategy["description"])
        this_layout.addRow("Description", description_input)
        risk_level_input = QtWidgets.QComboBox()
        red_pixmap = QtGui.QPixmap()
        red_pixmap.load(f"{introduction.PATH}/static/icon/traffic_light_red.png")
        yellow_pixmap = QtGui.QPixmap()
        yellow_pixmap.load(f"{introduction.PATH}/static/icon/traffic_light_yellow.png")
        green_pixmap = QtGui.QPixmap()
        green_pixmap.load(f"{introduction.PATH}/static/icon/traffic_light_green.png")
        risk_level_input.addItem(red_pixmap, "High")
        risk_level_input.addItem(yellow_pixmap, "Middle")
        risk_level_input.addItem(green_pixmap, "Low")
        risk_level_input.setCurrentIndex(strategy["risk_level"])
        this_layout.addRow("Risk level", risk_level_input)

        # ■■■■■ a card ■■■■■

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QVBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # explanation
        detail_text = QtWidgets.QLabel()
        detail_text.setText("Simulation")
        detail_text.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        detail_text.setWordWrap(True)
        card_layout.addWidget(detail_text)

        # spacing
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(3)
        spacing_text.setFont(spacing_text_font)
        card_layout.addWidget(spacing_text)

        # divider
        divider = HorizontalDivider(self)
        card_layout.addWidget(divider)

        # spacing
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(3)
        spacing_text.setFont(spacing_text_font)
        card_layout.addWidget(spacing_text)

        # input
        this_layout = QtWidgets.QFormLayout()
        card_layout.addLayout(this_layout)
        parallelized_input = QtWidgets.QCheckBox()
        parallelized_input.setChecked(strategy["parallelized_simulation"])
        this_layout.addRow("Parallelized", parallelized_input)
        chunk_division_input = QtWidgets.QSpinBox()
        chunk_division_input.setSuffix(" days")
        chunk_division_input.setMinimum(7)
        chunk_division_input.setMaximum(90)
        chunk_division_input.setButtonSymbols(
            QtWidgets.QSpinBox.ButtonSymbols.NoButtons
        )
        chunk_division_input.setValue(strategy["chunk_division"])
        this_layout.addRow("Chunk division", chunk_division_input)

        # ■■■■■ a card ■■■■■

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QHBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # function
        async def job(*args):
            code_name = code_name_input.text()
            if re.fullmatch(r"[A-Z]{6}", code_name):
                strategy["code_name"] = code_name
            else:
                question = [
                    "Code name format is wrong.",
                    "You should make the code name with 6 capital letters.",
                    ["Okay"],
                ]
                await solie.window.ask(question)
                return
            strategy["readable_name"] = readable_name_input.text()
            version = version_input.text()
            if re.fullmatch(r"[0-9]+\.[0-9]+", version):
                if not compare_versions.is_left_higher(strategy["version"], version):
                    strategy["version"] = version
                else:
                    question = [
                        "Version is lower.",
                        "You can't lower the version of your strategy. It should only"
                        " go up higher.",
                        ["Okay"],
                    ]
                    await solie.window.ask(question)
                    return
            else:
                question = [
                    "Version format is wrong.",
                    "You should write the version in two numeric fields, divided by a"
                    " single dot.",
                    ["Okay"],
                ]
                await solie.window.ask(question)
                return
            strategy["description"] = description_input.toPlainText()
            strategy["risk_level"] = risk_level_input.currentIndex()
            strategy["parallelized_simulation"] = parallelized_input.isChecked()
            strategy["chunk_division"] = chunk_division_input.value()
            done_event.set()

        # confirm button
        confirm_button = QtWidgets.QPushButton("Save and close", card)
        outsource.do(confirm_button.clicked, job)
        confirm_button.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )
        card_layout.addWidget(confirm_button)

        # ■■■■■ spacing ■■■■■

        # spacing
        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)



import asyncio

import aiofiles
from PySide6 import QtGui, QtWidgets

from solie import introduction
from solie.widget.brand_label import BrandLabel


class SplashScreen(QtWidgets.QFrame):
    def __init__(self):
        super().__init__()

        self.full_layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.full_layout)

        asyncio.create_task(self.fill())

    async def fill(self):
        full_layout = self.full_layout

        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )
        full_layout.addItem(spacer)

        central_layout = QtWidgets.QVBoxLayout()
        full_layout.addLayout(central_layout)

        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        central_layout.addItem(spacer)

        this_layout = QtWidgets.QHBoxLayout()
        central_layout.addLayout(this_layout)
        product_icon_pixmap = QtGui.QPixmap()
        async with aiofiles.open(
            f"{introduction.PATH}/static/product_icon.png", mode="rb"
        ) as file:
            product_icon_data = await file.read()
        product_icon_pixmap.loadFromData(product_icon_data)
        product_icon_label = QtWidgets.QLabel("", self)
        product_icon_label.setPixmap(product_icon_pixmap)
        product_icon_label.setScaledContents(True)
        product_icon_label.setFixedSize(80, 80)
        this_layout.addWidget(product_icon_label)
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(8)
        spacing_text.setFont(spacing_text_font)
        this_layout.addWidget(spacing_text)
        title_label = BrandLabel(self, "SOLIE", 48)
        this_layout.addWidget(title_label)
        text = introduction.VERSION
        label = BrandLabel(self, text, 24)
        this_layout.addWidget(label)

        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        central_layout.addItem(spacer)

        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )
        full_layout.addItem(spacer)



from datetime import datetime, timedelta, timezone

import pandas as pd


def do(
    symbol: str,
    recent_candle_data: pd.DataFrame,
    aggtrades: pd.DataFrame,
    moment_to_fill_from: datetime,
    last_fetched_time: datetime,
) -> pd.DataFrame:
    fill_moment = moment_to_fill_from

    while fill_moment < last_fetched_time - timedelta(seconds=10):
        block_start = fill_moment
        block_end = fill_moment + timedelta(seconds=10)

        aggtrade_prices = []
        aggtrade_volumes = []
        for _, aggtrade in sorted(aggtrades.items()):
            # sorted by time
            aggtrade_time = datetime.fromtimestamp(
                aggtrade["T"] / 1000, tz=timezone.utc
            )
            if block_start <= aggtrade_time < block_end:
                aggtrade_prices.append(float(aggtrade["p"]))
                aggtrade_volumes.append(float(aggtrade["q"]))

        can_write = True

        if len(aggtrade_prices) == 0:
            # when there are no trades
            inspect_sr = recent_candle_data[(symbol, "Close")]
            inspect_sr = inspect_sr.sort_index()
            last_prices = inspect_sr[:fill_moment].dropna()
            if len(last_prices) == 0:
                # when there are no previous data
                # because new data folder was created
                can_write = False
                last_price = 0
            else:
                last_price = last_prices.iloc[-1]
            open_price = last_price
            high_price = last_price
            low_price = last_price
            close_price = last_price
            sum_volume = 0
        else:
            open_price = aggtrade_prices[0]
            high_price = max(aggtrade_prices)
            low_price = min(aggtrade_prices)
            close_price = aggtrade_prices[-1]
            sum_volume = sum(aggtrade_volumes)

        if can_write:
            column = (symbol, "Open")
            recent_candle_data.loc[fill_moment, column] = open_price
            column = (symbol, "High")
            recent_candle_data.loc[fill_moment, column] = high_price
            column = (symbol, "Low")
            recent_candle_data.loc[fill_moment, column] = low_price
            column = (symbol, "Close")
            recent_candle_data.loc[fill_moment, column] = close_price
            column = (symbol, "Volume")
            recent_candle_data.loc[fill_moment, column] = sum_volume

        fill_moment += timedelta(seconds=10)

    return recent_candle_data



from collections import deque

task_durations = {
    "add_candle_data": deque(maxlen=360),
    "add_book_tickers": deque(maxlen=1280),
    "add_mark_price": deque(maxlen=10),
    "add_aggregate_trades": deque(maxlen=1280),
    "collector_organize_data": deque(maxlen=60),
    "perform_transaction": deque(maxlen=360),
    "display_light_transaction_lines": deque(maxlen=60),
    "display_all_transaction_lines": deque(maxlen=20),
    "place_orders": deque(maxlen=60),
}


def add(task_name, duration):
    task_durations[task_name].append(duration)


def get():
    return task_durations



import functools

import xdialog
from PySide6 import QtCore, QtGui, QtWidgets

import solie
from solie.recipe import outsource, user_settings
from solie.widget.horizontal_divider import HorizontalDivider


class DatapathInput(QtWidgets.QWidget):
    def __init__(self, done_event, payload):
        # ■■■■■ the basic ■■■■■

        super().__init__()

        # ■■■■■ full layout ■■■■■

        full_layout = QtWidgets.QHBoxLayout(self)
        cards_layout = QtWidgets.QVBoxLayout()
        full_layout.addLayout(cards_layout)

        # ■■■■■ spacing ■■■■■

        # spacing
        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)

        # ■■■■■ a card ■■■■■

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QVBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # explanation
        detail_text = QtWidgets.QLabel()
        detail_text.setText("All the data that Solie produces will go in this folder.")
        detail_text.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        detail_text.setWordWrap(True)
        card_layout.addWidget(detail_text)

        # spacing
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(3)
        spacing_text.setFont(spacing_text_font)
        card_layout.addWidget(spacing_text)

        # divider
        divider = HorizontalDivider(self)
        card_layout.addWidget(divider)

        # spacing
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(3)
        spacing_text.setFont(spacing_text_font)
        card_layout.addWidget(spacing_text)

        # chosen folder label
        folder_label = QtWidgets.QLabel()
        folder_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        folder_label.setWordWrap(True)
        card_layout.addWidget(folder_label)

        # spacing
        spacing_text = QtWidgets.QLabel("")
        spacing_text_font = QtGui.QFont()
        spacing_text_font.setPointSize(3)
        spacing_text.setFont(spacing_text_font)
        card_layout.addWidget(spacing_text)

        datapath = ""

        async def job_dp():
            nonlocal datapath

            datapath = await solie.event_loop.run_in_executor(
                None,
                functools.partial(
                    xdialog.directory,
                    title="Data folder",
                ),
            )

            folder_label.setText(datapath)

        # choose button
        this_layout = QtWidgets.QHBoxLayout()
        card_layout.addLayout(this_layout)
        choose_button = QtWidgets.QPushButton("Choose folder", card)
        outsource.do(choose_button.clicked, job_dp)
        choose_button.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )
        this_layout.addWidget(choose_button)

        # ■■■■■ a card ■■■■■

        async def job_ac():
            if datapath == "":
                question = [
                    "Data folder is not chosen",
                    "Choose your data folder first.",
                    ["Okay"],
                ]
                await solie.window.ask(question)
            else:
                await user_settings.apply_app_settings({"datapath": datapath})
                await user_settings.load()
                done_event.set()

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QHBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # confirm button
        confirm_button = QtWidgets.QPushButton("Okay", card)
        outsource.do(confirm_button.clicked, job_ac)
        confirm_button.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )
        card_layout.addWidget(confirm_button)

        # ■■■■■ spacing ■■■■■

        # spacing
        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)



import re


def numeric(string: str) -> str:
    string = re.sub(r"[^0123456789\.]", "", string)
    if string.startswith("."):
        string = "0" + string
    if string.endswith("."):
        string = string[:-1]
    return string


def fixed_float(number: int | float, width=4, positive_sign=False) -> str:
    if width < 4:
        width = 4

    if number < 0 or (positive_sign and number >= 0):
        # when sign should be included
        absolute_limit = 10 ** (width - 1)
    else:
        absolute_limit = 10**width

    if abs(number) >= absolute_limit:
        number = absolute_limit - 1 if number > 0 else -absolute_limit + 1

    string = f"{number:12.12f}"

    if positive_sign and number >= 0:
        string = "+" + string

    string = string[:width]

    return string



from pygments import lex, lexers
from PySide6 import QtGui


class SyntaxHighlighter(QtGui.QSyntaxHighlighter):
    def highlightBlock(self, script):  # noqa:N802
        current_position = 0

        for token, text_block in lex(script, self._lexer):
            length = len(text_block)
            text_format = None
            for text_type in self._mapping.keys():
                if str(token).startswith(text_type):
                    text_format = self._mapping[text_type]
                    break
            if text_format is not None:
                self.setFormat(current_position, length, text_format)
            current_position += length

    def __init__(self, parent):
        super().__init__(parent)

        self._lexer = lexers.PythonLexer()
        self._mapping = {}

        text_type = "Token.Keyword"
        hex_color = "#d1939e"
        text_format = QtGui.QTextCharFormat()
        text_format.setForeground(QtGui.QBrush(QtGui.QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Literal.String"
        hex_color = "#bce051"
        text_format = QtGui.QTextCharFormat()
        text_format.setForeground(QtGui.QBrush(QtGui.QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Literal.Number"
        hex_color = "#d1939e"
        text_format = QtGui.QTextCharFormat()
        text_format.setForeground(QtGui.QBrush(QtGui.QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Operator"
        hex_color = "#f4b73d"
        text_format = QtGui.QTextCharFormat()
        text_format.setForeground(QtGui.QBrush(QtGui.QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Punctuation"
        hex_color = "#999999"
        text_format = QtGui.QTextCharFormat()
        text_format.setForeground(QtGui.QBrush(QtGui.QColor(hex_color)))
        self._mapping[text_type] = text_format

        text_type = "Token.Comment"
        hex_color = "#997f66"
        text_format = QtGui.QTextCharFormat()
        text_format.setForeground(QtGui.QBrush(QtGui.QColor(hex_color)))
        self._mapping[text_type] = text_format



import asyncio
from typing import Callable, Coroutine

import aiohttp

import solie

is_ready = asyncio.Event()
_was_connected = False
_connected_functions: list[Callable[..., Coroutine]] = []
_disconnected_functions: list[Callable[..., Coroutine]] = []


def connected():
    if is_ready.is_set():
        return _was_connected
    else:
        raise RuntimeError("Internet connection is not being monitored")


async def monitor():
    global _was_connected
    while True:
        # try to connect to DNS servers
        attempt_ips = [
            "1.0.0.1",  # Cloudflare
            "1.1.1.1",  # Cloudflare
            "8.8.4.4",  # Google
            "8.8.8.8",  # Google
            "9.9.9.9",  # Quad9
            "149.112.112.112",  # Quad9
            "208.67.222.222",  # OpenDNS
            "208.67.220.220",  # OpenDNS
        ]
        is_connected = False
        async with aiohttp.ClientSession() as session:
            for attempt_ip in attempt_ips:
                try:
                    async with session.get(f"http://{attempt_ip}") as response:
                        if response.status == 200:
                            is_connected = True
                            break
                except aiohttp.ClientError as error:
                    solie.logger.exception(error)

        # detect changes
        if _was_connected and not is_connected:
            for job in _disconnected_functions:
                asyncio.create_task(job())
            solie.logger.warning("Internet disconnected")
        elif not _was_connected and is_connected:
            for job in _connected_functions:
                asyncio.create_task(job())
            solie.logger.info("Internet connected")

        # remember connection state
        _was_connected = is_connected
        is_ready.set()

        # wait for a while
        await asyncio.sleep(1)


def add_connected_functions(job: Callable[..., Coroutine]):
    global _connected_functions
    _connected_functions.append(job)


def add_disconnected_functions(job: Callable[..., Coroutine]):
    global _disconnected_functions
    _disconnected_functions.append(job)



import asyncio

from PySide6 import QtCore, QtGui, QtWidgets

from solie.recipe import outsource
from solie.widget.popup_box import PopupBox
from solie.widget.transparent_scroll_area import TransparentScrollArea

# https://stackoverflow.com/questions/67029993/pyqt-creating-a-popup-in-the-window


class OverlayPanel(QtWidgets.QWidget):
    done_event = asyncio.Event()

    def showEvent(self, event):  # noqa:N802
        # needed for filling the window when resized
        parent: QtWidgets.QMainWindow = self.parent()  # type:ignore
        self.setGeometry(parent.rect())

    def eventFilter(self, source, event):  # noqa:N802
        # needed for filling the window when resized
        if event.type() == event.Type.Resize:
            self.setGeometry(source.rect())
        return super().eventFilter(source, event)

    def __init__(self, parent, formation):
        # ■■■■■ the basic ■■■■■

        super().__init__(parent)

        # ■■■■■ set properties ■■■■■

        # needed for filling the window when resized
        parent.installEventFilter(self)

        # ■■■■■ in case other ask popup exists ■■■■■

        self.done_event.set()
        self.done_event.clear()

        # ■■■■■ prepare answer ■■■■■

        self.answer = 0

        # ■■■■■ full structure ■■■■■

        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_StyledBackground)
        self.setAutoFillBackground(True)

        # ■■■■■ full layout ■■■■■

        full_layout = QtWidgets.QHBoxLayout(self)

        # ■■■■■ visaul box ■■■■■

        # box
        content_box = PopupBox()
        content_box.setAutoFillBackground(True)
        content_box.setMaximumSize(1600, 1200)
        content_box_layout = QtWidgets.QVBoxLayout(content_box)
        full_layout.addWidget(content_box)

        # line containing the title and close button
        this_layout = QtWidgets.QHBoxLayout()
        title_label = QtWidgets.QLabel(formation[0])
        title_label_font = QtGui.QFont()
        title_label_font.setPointSize(12)
        title_label.setFont(title_label_font)
        title_label.setWordWrap(False)
        this_layout.addWidget(title_label)
        widget = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )
        this_layout.addItem(widget)
        if formation[2]:
            close_button = QtWidgets.QPushButton("✕", content_box)
            close_button_font = QtGui.QFont()
            close_button_font.setPointSize(11)
            close_button.setFont(close_button_font)

            async def job():
                self.done_event.set()

            outsource.do(close_button.clicked, job)
            this_layout.addWidget(close_button)
        content_box_layout.addLayout(this_layout)

        # scroll area
        scroll_area = TransparentScrollArea()
        scroll_widget = formation[1](self.done_event, formation[3])

        scroll_area.setWidget(scroll_widget)
        scroll_area.setWidgetResizable(True)

        content_box_layout.addWidget(scroll_area)



import asyncio
from typing import Callable, Coroutine

from PySide6.QtCore import SignalInstance


def do(signal: SignalInstance, target_function: Callable[..., Coroutine]):
    def job(*args):
        asyncio.create_task(target_function(*args))

    signal.connect(job)



from PySide6 import QtWidgets


class HorizontalDivider(QtWidgets.QFrame):
    def __init__(self, parent):
        super().__init__(parent)
        self.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.setFixedHeight(2)



import asyncio
import functools
import math
import os
import pickle
import re
from datetime import datetime, timedelta, timezone

import aiofiles
import numpy as np
import pandas as pd
from scipy.signal import find_peaks

import solie
from solie.definition.rw_lock import RWLock
from solie.recipe import (
    make_indicators,
    simulate_chunk,
    standardize,
    stop_flag,
    user_settings,
)


class Simulator:
    def __init__(self):
        # ■■■■■ for data management ■■■■■

        self.workerpath = user_settings.get_app_settings()["datapath"] + "/simulator"
        os.makedirs(self.workerpath, exist_ok=True)

        # ■■■■■ worker secret memory ■■■■■

        self.secret_memory = {}

        # ■■■■■ remember and display ■■■■■

        self.viewing_symbol = user_settings.get_data_settings()["target_symbols"][0]
        self.should_draw_all_years = False

        self.about_viewing = None

        self.calculation_settings = {
            "year": datetime.now(timezone.utc).year,
            "strategy_index": 0,
        }
        self.presentation_settings = {
            "maker_fee": 0.02,
            "taker_fee": 0.04,
            "leverage": 1,
        }

        self.raw_account_state = standardize.account_state()
        self.raw_scribbles = {}
        self.raw_asset_record = RWLock(standardize.asset_record())
        self.raw_unrealized_changes = RWLock(standardize.unrealized_changes())

        self.account_state = standardize.account_state()
        self.scribbles = {}
        self.asset_record = RWLock(standardize.asset_record())
        self.unrealized_changes = RWLock(standardize.unrealized_changes())

        # ■■■■■ repetitive schedules ■■■■■

        solie.window.scheduler.add_job(
            self.display_available_years,
            trigger="cron",
            second="*",
        )
        solie.window.scheduler.add_job(
            self.display_lines,
            trigger="cron",
            hour="*",
            kwargs={"periodic": True},
        )

        # ■■■■■ websocket streamings ■■■■■

        self.api_streamers = {}

        # ■■■■■ invoked by the internet connection status change  ■■■■■

    async def load(self, *args, **kwargs):
        text = "Nothing drawn"
        solie.window.label_19.setText(text)

    async def update_viewing_symbol(self, *args, **kwargs):
        alias = solie.window.comboBox_6.currentText()
        symbol = solie.window.alias_to_symbol[alias]
        self.viewing_symbol = symbol

        await self.display_lines()

    async def update_calculation_settings(self, *args, **kwargs):
        text = solie.window.comboBox_5.currentText()
        if text == "":
            return
        from_year = self.calculation_settings["year"]
        to_year = int(text)
        self.calculation_settings["year"] = to_year
        if from_year != to_year:
            asyncio.create_task(self.display_year_range())

        index = solie.window.comboBox.currentIndex()
        self.calculation_settings["strategy_index"] = index

        await self.display_lines()

    async def update_presentation_settings(self, *args, **kwargs):
        input_value = solie.window.spinBox_2.value()
        self.presentation_settings["leverage"] = input_value
        input_value = solie.window.doubleSpinBox.value()
        self.presentation_settings["taker_fee"] = input_value
        input_value = solie.window.doubleSpinBox_2.value()
        self.presentation_settings["maker_fee"] = input_value
        await self.present()

    async def display_lines(self, *args, **kwargs):
        # ■■■■■ start the task ■■■■■

        periodic = kwargs.get("periodic", False)
        frequent = kwargs.get("frequent", False)
        only_light_lines = kwargs.get("only_light_lines", False)

        if only_light_lines:
            task_name = "display_light_simulation_lines"
        else:
            task_name = "display_all_simulation_lines"

        task_id = stop_flag.make(task_name)

        # ■■■■■ check drawing mode ■■■■■

        should_draw_all_years = self.should_draw_all_years

        if frequent:
            pass

        # ■■■■■ check if the data exists ■■■■■

        async with solie.window.collector.candle_data.read_lock as wrapper:
            inner = wrapper.inner
            if len(inner) == 0:
                return

        # ■■■■■ wait for the latest data to be added ■■■■■

        current_moment = datetime.now(timezone.utc).replace(microsecond=0)
        current_moment = current_moment - timedelta(seconds=current_moment.second % 10)
        before_moment = current_moment - timedelta(seconds=10)

        if periodic:
            for _ in range(50):
                if stop_flag.find(task_name, task_id):
                    return
                async with solie.window.collector.candle_data.read_lock as wrapper:
                    inner = wrapper.inner
                    last_index = inner.index[-1]
                    if last_index == before_moment:
                        break
                await asyncio.sleep(0.1)

        # ■■■■■ get ready for task duration measurement ■■■■■

        pass

        # ■■■■■ check things ■■■■■

        symbol = self.viewing_symbol
        strategy_index = self.calculation_settings["strategy_index"]
        strategy = solie.window.strategist.strategies[strategy_index]

        # ■■■■■ get light data ■■■■■

        async with solie.window.collector.realtime_data_chunks.read_lock as wrapper:
            inner = wrapper.inner
            before_chunk = inner[-2].copy()
            current_chunk = inner[-1].copy()
        realtime_data = np.concatenate((before_chunk, current_chunk))
        async with solie.window.collector.aggregate_trades.read_lock as wrapper:
            aggregate_trades = wrapper.inner.copy()

        # ■■■■■ draw light lines ■■■■■

        # mark price
        data_x = realtime_data["index"].astype(np.int64) / 10**9
        data_y = realtime_data[str((symbol, "Mark Price"))]
        mask = data_y != 0
        data_y = data_y[mask]
        data_x = data_x[mask]
        widget = solie.window.simulation_lines["mark_price"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # last price
        data_x = aggregate_trades["index"].astype(np.int64) / 10**9
        data_y = aggregate_trades[str((symbol, "Price"))]
        mask = data_y != 0
        data_y = data_y[mask]
        data_x = data_x[mask]
        widget = solie.window.simulation_lines["last_price"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # last trade volume
        index_ar = aggregate_trades["index"].astype(np.int64) / 10**9
        value_ar = aggregate_trades[str((symbol, "Volume"))]
        mask = value_ar != 0
        index_ar = index_ar[mask]
        value_ar = value_ar[mask]
        length = len(index_ar)
        zero_ar = np.zeros(length)
        nan_ar = np.empty(length)
        nan_ar[:] = np.nan
        data_x = np.repeat(index_ar, 3)
        data_y = np.stack([nan_ar, zero_ar, value_ar], axis=1).reshape(-1)
        widget = solie.window.simulation_lines["last_volume"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # book tickers
        data_x = realtime_data["index"].astype(np.int64) / 10**9
        data_y = realtime_data[str((symbol, "Best Bid Price"))]
        mask = data_y != 0
        data_y = data_y[mask]
        data_x = data_x[mask]
        widget = solie.window.simulation_lines["book_tickers"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        data_x = realtime_data["index"].astype(np.int64) / 10**9
        data_y = realtime_data[str((symbol, "Best Ask Price"))]
        mask = data_y != 0
        data_y = data_y[mask]
        data_x = data_x[mask]
        widget = solie.window.simulation_lines["book_tickers"][1]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # entry price
        entry_price = self.account_state["positions"][symbol]["entry_price"]
        first_moment = self.account_state["observed_until"] - timedelta(hours=12)
        last_moment = self.account_state["observed_until"] + timedelta(hours=12)
        if entry_price != 0:
            data_x = np.linspace(
                first_moment.timestamp(), last_moment.timestamp(), num=1000
            )
            data_y = np.linspace(entry_price, entry_price, num=1000)
        else:
            data_x = []
            data_y = []
        widget = solie.window.simulation_lines["entry_price"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # ■■■■■ record task duration ■■■■■

        pass

        # ■■■■■ stop if the target is only light lines ■■■■■

        if only_light_lines:
            return

        # ■■■■■ set range of heavy data ■■■■■

        if should_draw_all_years:
            slice_from = datetime.fromtimestamp(0, tz=timezone.utc)
            slice_until = datetime.now(timezone.utc)
            slice_until = slice_until.replace(minute=0, second=0, microsecond=0)
        else:
            year = self.calculation_settings["year"]
            slice_from = datetime(year, 1, 1, tzinfo=timezone.utc)
            if year == datetime.now(timezone.utc).year:
                slice_until = datetime.now(timezone.utc)
                slice_until = slice_until.replace(minute=0, second=0, microsecond=0)
            else:
                slice_until = datetime(year + 1, 1, 1, tzinfo=timezone.utc)
        slice_until -= timedelta(seconds=1)
        get_from = slice_from - timedelta(days=7)

        # ■■■■■ get heavy data ■■■■■

        async with solie.window.collector.candle_data.read_lock as wrapper:
            inner = wrapper.inner
            inner = inner[get_from:slice_until][[symbol]]
            candle_data = inner.copy()
        async with self.unrealized_changes.read_lock as wrapper:
            unrealized_changes = wrapper.inner.copy()
        async with self.asset_record.read_lock as wrapper:
            inner = wrapper.inner
            if len(inner) > 0:
                last_asset = inner.iloc[-1]["Result Asset"]
            else:
                last_asset = None
            before_record = inner[:slice_from]
            inner = inner[slice_from:]
            if len(before_record) > 0:
                before_asset = before_record.iloc[-1]["Result Asset"]
            else:
                before_asset = None
            asset_record = inner.copy()

        candle_data = candle_data[slice_from:]

        # ■■■■■ maniuplate heavy data ■■■■■

        # add the right end

        if len(candle_data) > 0:
            last_written_moment = candle_data.index[-1]
            new_moment = last_written_moment + timedelta(seconds=10)
            new_index = candle_data.index.union([new_moment])
            candle_data = candle_data.reindex(new_index)

        if last_asset is not None:
            observed_until = self.account_state["observed_until"]
            if len(asset_record) == 0 or asset_record.index[-1] < observed_until:
                if slice_from < observed_until:
                    asset_record.loc[observed_until, "Cause"] = "other"
                    asset_record.loc[observed_until, "Result Asset"] = last_asset
                    asset_record = asset_record.sort_index()

        # add the left end

        if before_asset is not None:
            asset_record.loc[slice_from, "Cause"] = "other"
            asset_record.loc[slice_from, "Result Asset"] = before_asset
            asset_record = asset_record.sort_index()

        # ■■■■■ draw heavy lines ■■■■■

        # price movement
        index_ar = candle_data.index.to_numpy(dtype=np.int64) / 10**9
        open_ar = candle_data[(symbol, "Open")].to_numpy()
        close_ar = candle_data[(symbol, "Close")].to_numpy()
        high_ar = candle_data[(symbol, "High")].to_numpy()
        low_ar = candle_data[(symbol, "Low")].to_numpy()
        rise_ar = close_ar > open_ar
        fall_ar = close_ar < open_ar
        stay_ar = close_ar == open_ar
        length = len(index_ar)
        nan_ar = np.empty(length)
        nan_ar[:] = np.nan

        data_x = np.stack(
            [
                index_ar[rise_ar] + 2,
                index_ar[rise_ar] + 5,
                index_ar[rise_ar],
                index_ar[rise_ar] + 5,
                index_ar[rise_ar] + 8,
                index_ar[rise_ar],
                index_ar[rise_ar] + 5,
                index_ar[rise_ar] + 5,
                index_ar[rise_ar],
            ],
            axis=1,
        ).reshape(-1)
        data_y = np.stack(
            [
                open_ar[rise_ar],
                open_ar[rise_ar],
                nan_ar[rise_ar],
                close_ar[rise_ar],
                close_ar[rise_ar],
                nan_ar[rise_ar],
                high_ar[rise_ar],
                low_ar[rise_ar],
                nan_ar[rise_ar],
            ],
            axis=1,
        ).reshape(-1)
        widget = solie.window.simulation_lines["price_rise"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        data_x = np.stack(
            [
                index_ar[fall_ar] + 2,
                index_ar[fall_ar] + 5,
                index_ar[fall_ar],
                index_ar[fall_ar] + 5,
                index_ar[fall_ar] + 8,
                index_ar[fall_ar],
                index_ar[fall_ar] + 5,
                index_ar[fall_ar] + 5,
                index_ar[fall_ar],
            ],
            axis=1,
        ).reshape(-1)
        data_y = np.stack(
            [
                open_ar[fall_ar],
                open_ar[fall_ar],
                nan_ar[fall_ar],
                close_ar[fall_ar],
                close_ar[fall_ar],
                nan_ar[fall_ar],
                high_ar[fall_ar],
                low_ar[fall_ar],
                nan_ar[fall_ar],
            ],
            axis=1,
        ).reshape(-1)
        widget = solie.window.simulation_lines["price_fall"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        data_x = np.stack(
            [
                index_ar[stay_ar] + 2,
                index_ar[stay_ar] + 5,
                index_ar[stay_ar],
                index_ar[stay_ar] + 5,
                index_ar[stay_ar] + 8,
                index_ar[stay_ar],
                index_ar[stay_ar] + 5,
                index_ar[stay_ar] + 5,
                index_ar[stay_ar],
            ],
            axis=1,
        ).reshape(-1)
        data_y = np.stack(
            [
                open_ar[stay_ar],
                open_ar[stay_ar],
                nan_ar[stay_ar],
                close_ar[stay_ar],
                close_ar[stay_ar],
                nan_ar[stay_ar],
                high_ar[stay_ar],
                low_ar[stay_ar],
                nan_ar[stay_ar],
            ],
            axis=1,
        ).reshape(-1)
        widget = solie.window.simulation_lines["price_stay"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # wobbles
        sr = candle_data[(symbol, "High")]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        widget = solie.window.simulation_lines["wobbles"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        sr = candle_data[(symbol, "Low")]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        widget = solie.window.simulation_lines["wobbles"][1]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # trade volume
        sr = candle_data[(symbol, "Volume")]
        sr = sr.fillna(value=0)
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        widget = solie.window.simulation_lines["volume"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # asset
        data_x = asset_record["Result Asset"].index.to_numpy(dtype=np.int64) / 10**9
        data_y = asset_record["Result Asset"].to_numpy(dtype=np.float32)
        widget = solie.window.simulation_lines["asset"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # asset with unrealized profit
        sr = asset_record["Result Asset"]
        if len(sr) >= 2:
            sr = sr.resample("10S").ffill()
        unrealized_changes_sr = unrealized_changes.reindex(sr.index)
        sr = sr * (1 + unrealized_changes_sr)
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9 + 5
        data_y = sr.to_numpy(dtype=np.float32)
        widget = solie.window.simulation_lines["asset_with_unrealized_profit"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # buy and sell
        df = asset_record.loc[asset_record["Symbol"] == symbol]
        df = df[df["Side"] == "sell"]
        sr = df["Fill Price"]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        widget = solie.window.simulation_lines["sell"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        df = asset_record.loc[asset_record["Symbol"] == symbol]
        df = df[df["Side"] == "buy"]
        sr = df["Fill Price"]
        data_x = sr.index.to_numpy(dtype=np.int64) / 10**9
        data_y = sr.to_numpy(dtype=np.float32)
        widget = solie.window.simulation_lines["buy"][0]
        widget.setData(data_x, data_y)
        if stop_flag.find(task_name, task_id):
            return
        await asyncio.sleep(0)

        # ■■■■■ record task duration ■■■■■

        pass

        # ■■■■■ make indicators ■■■■■

        indicators_script = strategy["indicators_script"]

        indicators = await solie.event_loop.run_in_executor(
            solie.process_pool,
            functools.partial(
                make_indicators.do,
                target_symbols=[self.viewing_symbol],
                candle_data=candle_data,
                indicators_script=indicators_script,
            ),
        )

        indicators = indicators[slice_from:]

        # ■■■■■ draw strategy lines ■■■■■

        # price indicators
        df = indicators[symbol]["Price"]
        data_x = df.index.to_numpy(dtype=np.int64) / 10**9
        data_x += 5
        line_list = solie.window.simulation_lines["price_indicators"]
        for turn, widget in enumerate(line_list):
            if turn < len(df.columns):
                column_name = df.columns[turn]
                sr = df[column_name]
                data_y = sr.to_numpy(dtype=np.float32)
                inside_strings = re.findall(r"\(([^)]+)", column_name)
                if len(inside_strings) == 0:
                    color = "#AAAAAA"
                else:
                    color = inside_strings[0]
                widget.setPen(color)
                widget.setData(data_x, data_y)
                if stop_flag.find(task_name, task_id):
                    return
                await asyncio.sleep(0)
            else:
                if stop_flag.find(task_name, task_id):
                    return
                widget.clear()

        # trade volume indicators
        df = indicators[symbol]["Volume"]
        data_x = df.index.to_numpy(dtype=np.int64) / 10**9
        data_x += 5
        line_list = solie.window.simulation_lines["volume_indicators"]
        for turn, widget in enumerate(line_list):
            if turn < len(df.columns):
                column_name = df.columns[turn]
                sr = df[column_name]
                data_y = sr.to_numpy(dtype=np.float32)
                inside_strings = re.findall(r"\(([^)]+)", column_name)
                if len(inside_strings) == 0:
                    color = "#AAAAAA"
                else:
                    color = inside_strings[0]
                widget.setPen(color)
                widget.setData(data_x, data_y)
                if stop_flag.find(task_name, task_id):
                    return
                await asyncio.sleep(0)
            else:
                if stop_flag.find(task_name, task_id):
                    return
                widget.clear()

        # abstract indicators
        df = indicators[symbol]["Abstract"]
        data_x = df.index.to_numpy(dtype=np.int64) / 10**9
        data_x += 5
        line_list = solie.window.simulation_lines["abstract_indicators"]
        for turn, widget in enumerate(line_list):
            if turn < len(df.columns):
                column_name = df.columns[turn]
                sr = df[column_name]
                data_y = sr.to_numpy(dtype=np.float32)
                inside_strings = re.findall(r"\(([^)]+)", column_name)
                if len(inside_strings) == 0:
                    color = "#AAAAAA"
                else:
                    color = inside_strings[0]
                widget.setPen(color)
                widget.setData(data_x, data_y)
                if stop_flag.find(task_name, task_id):
                    return
                await asyncio.sleep(0)
            else:
                if stop_flag.find(task_name, task_id):
                    return
                widget.clear()

        # ■■■■■ set minimum view range ■■■■■

        await self.set_minimum_view_range()

    async def erase(self, *args, **kwargs):
        self.raw_account_state = standardize.account_state()
        self.raw_scribbles = {}
        self.raw_asset_record = RWLock(standardize.asset_record())
        self.raw_unrealized_changes = RWLock(standardize.unrealized_changes())
        self.about_viewing = None

        await self.present()

    async def display_available_years(self, *args, **kwargs):
        async with solie.window.collector.candle_data.read_lock as wrapper:
            inner = wrapper.inner
            years_sr = inner.index.year.drop_duplicates()  # type:ignore
        years = years_sr.tolist()
        years.sort(reverse=True)
        years = [str(year) for year in years]

        widget = solie.window.comboBox_5
        choices = [int(widget.itemText(i)) for i in range(widget.count())]
        choices.sort(reverse=True)
        choices = [str(choice) for choice in choices]

        if years != choices:
            # if it's changed
            solie.window.comboBox_5.clear()
            solie.window.comboBox_5.addItems(years)

    async def simulate_only_visible(self, *args, **kwargs):
        await self.calculate(only_visible=True)

    async def display_range_information(self, *args, **kwargs):
        task_id = stop_flag.make("display_simulation_range_information")

        symbol = self.viewing_symbol

        range_start_timestamp = solie.window.plot_widget_2.getAxis("bottom").range[0]
        range_start_timestamp = max(range_start_timestamp, 0.0)
        range_start = datetime.fromtimestamp(range_start_timestamp, tz=timezone.utc)

        if stop_flag.find("display_simulation_range_information", task_id):
            return

        range_end_timestamp = solie.window.plot_widget_2.getAxis("bottom").range[1]
        if range_end_timestamp < 0:
            # case when pyqtgraph passed negative value because it's too big
            range_end_timestamp = 9223339636
        else:
            # maximum value available in pandas
            range_end_timestamp = min(range_end_timestamp, 9223339636.0)
        range_end = datetime.fromtimestamp(range_end_timestamp, tz=timezone.utc)

        if stop_flag.find("display_simulation_range_information", task_id):
            return

        range_length = range_end - range_start
        range_days = range_length.days
        range_hours, remains = divmod(range_length.seconds, 3600)
        range_minutes, remains = divmod(remains, 60)

        if stop_flag.find("display_simulation_range_information", task_id):
            return

        async with self.unrealized_changes.read_lock as wrapper:
            unrealized_changes = wrapper.inner[range_start:range_end].copy()
        async with self.asset_record.read_lock as wrapper:
            asset_record = wrapper.inner[range_start:range_end].copy()

        auto_trade_mask = asset_record["Cause"] == "auto_trade"
        asset_changes = asset_record["Result Asset"].pct_change() + 1
        asset_record = asset_record[auto_trade_mask]
        asset_changes = asset_changes.reindex(asset_record.index).fillna(value=1)
        symbol_mask = asset_record["Symbol"] == symbol

        # trade count
        total_change_count = len(asset_changes)
        symbol_change_count = len(asset_changes[symbol_mask])
        # trade volume
        if len(asset_record) > 0:
            total_margin_ratio = asset_record["Margin Ratio"].sum()
        else:
            total_margin_ratio = 0
        if len(asset_record[symbol_mask]) > 0:
            symbol_margin_ratio = asset_record[symbol_mask]["Margin Ratio"].sum()
        else:
            symbol_margin_ratio = 0
        # asset changes
        if len(asset_changes) > 0:
            total_yield = asset_changes.cumprod().iloc[-1]
            total_yield = (total_yield - 1) * 100
        else:
            total_yield = 0
        if len(asset_changes[symbol_mask]) > 0:
            symbol_yield = asset_changes[symbol_mask].cumprod().iloc[-1]
            symbol_yield = (symbol_yield - 1) * 100
        else:
            symbol_yield = 0
        # least unrealized changes
        if len(unrealized_changes) > 0:
            min_unrealized_change = unrealized_changes.min()
        else:
            min_unrealized_change = 0

        if stop_flag.find("display_simulation_range_information", task_id):
            return

        view_range = solie.window.plot_widget_2.getAxis("left").range
        range_down = view_range[0]
        range_up = view_range[1]
        price_range_height = (1 - range_down / range_up) * 100

        text = ""
        text += f"Visible time range {range_days}d {range_hours}h {range_minutes}s"
        text += "  ⦁  "
        text += "Visible price range"
        text += f" {price_range_height:.2f}%"
        text += "  ⦁  "
        text += f"Transaction count {symbol_change_count}({total_change_count})"
        text += "  ⦁  "
        text += "Transaction amount"
        text += f" ×{symbol_margin_ratio:.4f}({total_margin_ratio:.4f})"
        text += "  ⦁  "
        text += f"Total realized profit {symbol_yield:+.4f}({total_yield:+.4f})%"
        text += "  ⦁  "
        text += "Lowest unrealized profit"
        text += f" {min_unrealized_change*100:+.4f}%"
        solie.window.label_13.setText(text)

    async def set_minimum_view_range(self, *args, **kwargs):
        widget = solie.window.plot_widget_2
        range_down = widget.getAxis("left").range[0]
        widget.plotItem.vb.setLimits(minYRange=range_down * 0.005)  # type:ignore
        widget = solie.window.plot_widget_3
        range_down = widget.getAxis("left").range[0]
        widget.plotItem.vb.setLimits(minYRange=range_down * 0.005)  # type:ignore

    async def calculate(self, *args, **kwargs):
        task_id = stop_flag.make("calculate_simulation")

        only_visible = kwargs.get("only_visible", False)

        prepare_step = 0
        calculate_step = 0

        async def play_progress_bar():
            while True:
                if stop_flag.find("calculate_simulation", task_id):
                    solie.window.progressBar_4.setValue(0)
                    solie.window.progressBar.setValue(0)
                    return
                else:
                    if prepare_step == 6 and calculate_step == 1000:
                        is_progressbar_filled = True
                        progressbar_value = solie.window.progressBar_4.value()
                        if progressbar_value < 1000:
                            is_progressbar_filled = False
                        progressbar_value = solie.window.progressBar.value()
                        if progressbar_value < 1000:
                            is_progressbar_filled = False
                        if is_progressbar_filled:
                            await asyncio.sleep(0.1)
                            solie.window.progressBar_4.setValue(0)
                            solie.window.progressBar.setValue(0)
                            return
                    widget = solie.window.progressBar_4
                    before_value = widget.value()
                    if before_value < 1000:
                        remaining = math.ceil(1000 / 6 * prepare_step) - before_value
                        new_value = before_value + math.ceil(remaining * 0.2)
                        widget.setValue(new_value)
                    widget = solie.window.progressBar
                    before_value = widget.value()
                    if before_value < 1000:
                        remaining = calculate_step - before_value
                        new_value = before_value + math.ceil(remaining * 0.2)
                        widget.setValue(new_value)
                    await asyncio.sleep(0.01)

        asyncio.create_task(play_progress_bar())

        prepare_step = 1

        # ■■■■■ default values and the strategy ■■■■■

        year = self.calculation_settings["year"]
        strategy_index = self.calculation_settings["strategy_index"]

        strategy = solie.window.strategist.strategies[strategy_index]
        strategy_code_name = strategy["code_name"]
        strategy_version = strategy["version"]
        should_parallelize = strategy["parallelized_simulation"]
        chunk_length = strategy["chunk_division"]

        path_start = f"{self.workerpath}/{strategy_code_name}_{strategy_version}_{year}"
        asset_record_path = path_start + "_asset_record.pickle"
        unrealized_changes_path = path_start + "_unrealized_changes.pickle"
        scribbles_path = path_start + "_scribbles.pickle"
        account_state_path = path_start + "_account_state.pickle"
        virtual_state_path = path_start + "_virtual_state.pickle"

        target_symbols = user_settings.get_data_settings()["target_symbols"]

        prepare_step = 2

        # ■■■■■ get datas ■■■■■

        # set range
        if self.should_draw_all_years:
            slice_from = datetime(1970, 1, 1, tzinfo=timezone.utc)
        else:
            slice_from = datetime(year, 1, 1, tzinfo=timezone.utc)

        if self.should_draw_all_years:
            slice_until = datetime(2200, 1, 1, tzinfo=timezone.utc)
        elif year == datetime.now(timezone.utc).year:
            slice_until = datetime.now(timezone.utc)
            slice_until = slice_until.replace(minute=0, second=0, microsecond=0)
        else:
            slice_until = datetime(year + 1, 1, 1, tzinfo=timezone.utc)
        slice_until -= timedelta(seconds=1)

        # get only year range
        async with solie.window.collector.candle_data.read_lock as wrapper:
            df = wrapper.inner
            year_candle_data = df[slice_from - timedelta(days=7) : slice_until].copy()

        # interpolate
        year_candle_data = year_candle_data.interpolate()

        prepare_step = 3

        # ■■■■■ prepare data and calculation range ■■■■■

        blank_asset_record = standardize.asset_record()
        blank_unrealized_changes = standardize.unrealized_changes()
        blank_scribbles = {}
        blank_account_state = standardize.account_state()
        blank_virtual_state = {
            "available_balance": 1,
            "locations": {},
            "placements": {},
        }
        for symbol in target_symbols:
            blank_virtual_state["locations"][symbol] = {
                "amount": 0,
                "entry_price": 0,
            }
            blank_virtual_state["placements"][symbol] = {}

        prepare_step = 4

        if only_visible:
            # when calculating only on visible range

            previous_asset_record = blank_asset_record.copy()
            previous_unrealized_changes = blank_unrealized_changes.copy()
            previous_scribbles = blank_scribbles.copy()
            previous_account_state = blank_account_state.copy()
            previous_virtual_state = blank_virtual_state.copy()

            view_range = solie.window.plot_widget_2.getAxis("bottom").range
            view_start = datetime.fromtimestamp(view_range[0], tz=timezone.utc)
            view_end = datetime.fromtimestamp(view_range[1], tz=timezone.utc)

            if self.should_draw_all_years:
                calculate_from = view_start
                calculate_until = view_end
            else:
                calculate_from = max(view_start, slice_from)
                calculate_until = min(view_end, slice_until)

        else:
            # when calculating properly
            try:
                filepath = asset_record_path
                previous_asset_record = pd.read_pickle(filepath)
                filepath = unrealized_changes_path
                previous_unrealized_changes = pd.read_pickle(filepath)
                async with aiofiles.open(scribbles_path, "rb") as file:
                    content = await file.read()
                    previous_scribbles = pickle.loads(content)
                async with aiofiles.open(account_state_path, "rb") as file:
                    content = await file.read()
                    previous_account_state = pickle.loads(content)
                async with aiofiles.open(virtual_state_path, "rb") as file:
                    content = await file.read()
                    previous_virtual_state = pickle.loads(content)

                calculate_from = previous_account_state["observed_until"]
                calculate_until = slice_until
            except FileNotFoundError:
                previous_asset_record = blank_asset_record.copy()
                previous_unrealized_changes = blank_unrealized_changes.copy()
                previous_scribbles = blank_scribbles.copy()
                previous_account_state = blank_account_state.copy()
                previous_virtual_state = blank_virtual_state.copy()

                calculate_from = slice_from
                calculate_until = slice_until

        should_calculate = calculate_from < calculate_until
        if len(previous_asset_record) == 0:
            previous_asset_record.loc[calculate_from, "Cause"] = "other"
            previous_asset_record.loc[calculate_from, "Result Asset"] = float(1)

        prepare_step = 5

        # ■■■■■ prepare per chunk data ■■■■■

        calculation_input_data = []
        progress_list = solie.communicator.list([0])

        if should_calculate:
            decision_script = strategy["decision_script"]
            indicators_script = strategy["indicators_script"]

            # a little more data for generation
            provide_from = calculate_from - timedelta(days=7)
            year_indicators = await solie.event_loop.run_in_executor(
                solie.process_pool,
                functools.partial(
                    make_indicators.do,
                    target_symbols=target_symbols,
                    candle_data=year_candle_data[provide_from:calculate_until],
                    indicators_script=indicators_script,
                ),
            )

            # range cut
            needed_candle_data = year_candle_data[calculate_from:calculate_until]
            needed_index = needed_candle_data.index
            needed_indicators = year_indicators.reindex(needed_index)

            if should_parallelize:
                division = timedelta(days=chunk_length)
                chunk_candle_data_list = [
                    chunk_candle_data
                    for _, chunk_candle_data in needed_candle_data.groupby(
                        pd.Grouper(freq=division, origin="epoch")  # type:ignore
                    )
                ]

                chunk_count = len(chunk_candle_data_list)
                progress_list = solie.communicator.list([0] * chunk_count)

                for turn, chunk_candle_data in enumerate(chunk_candle_data_list):
                    chunk_index = chunk_candle_data.index
                    chunk_indicators = needed_indicators.reindex(chunk_index)
                    chunk_asset_record = previous_asset_record.iloc[0:0]
                    chunk_unrealized_changes = previous_unrealized_changes.iloc[0:0]
                    first_timestamp = chunk_index[0].timestamp()
                    division_seconds = chunk_length * 24 * 60 * 60
                    if turn == 0 and first_timestamp % division_seconds != 0:
                        # when this is the firstmost chunk of calculation
                        # and also chunk calculation was partially done before
                        chunk_scribbles = previous_scribbles
                        chunk_account_state = previous_account_state
                        chunk_virtual_state = previous_virtual_state
                    else:
                        chunk_scribbles = blank_scribbles
                        chunk_account_state = blank_account_state
                        chunk_virtual_state = blank_virtual_state

                    dataset = {
                        "progress_list": progress_list,
                        "target_progress": turn,
                        "target_symbols": target_symbols,
                        "calculation_index": chunk_index,
                        "chunk_candle_data": chunk_candle_data,
                        "chunk_indicators": chunk_indicators,
                        "chunk_asset_record": chunk_asset_record,
                        "chunk_unrealized_changes": chunk_unrealized_changes,
                        "chunk_scribbles": chunk_scribbles,
                        "chunk_account_state": chunk_account_state,
                        "chunk_virtual_state": chunk_virtual_state,
                        "decision_script": decision_script,
                    }
                    calculation_input_data.append(dataset)

            else:
                dataset = {
                    "progress_list": progress_list,
                    "target_progress": 0,
                    "target_symbols": target_symbols,
                    "calculation_index": needed_index,
                    "chunk_candle_data": needed_candle_data,
                    "chunk_indicators": needed_indicators,
                    "chunk_asset_record": previous_asset_record,
                    "chunk_unrealized_changes": previous_unrealized_changes,
                    "chunk_scribbles": previous_scribbles,
                    "chunk_account_state": previous_account_state,
                    "chunk_virtual_state": previous_virtual_state,
                    "decision_script": decision_script,
                }
                calculation_input_data.append(dataset)

        prepare_step = 6

        # ■■■■■ calculate ■■■■■

        calculation_output_data = []

        if should_calculate:
            futures = [
                solie.event_loop.run_in_executor(
                    solie.process_pool, simulate_chunk.do, input_data
                )
                for input_data in calculation_input_data
            ]
            gathered = asyncio.gather(*futures)

            total_seconds = (calculate_until - calculate_from).total_seconds()

            async def update_calculation_step():
                nonlocal calculate_step
                while True:
                    if stop_flag.find("calculate_simulation", task_id):
                        return
                    if gathered.done():
                        return
                    total_progress = sum(progress_list)
                    calculate_step = math.ceil(total_progress * 1000 / total_seconds)
                    await asyncio.sleep(0.01)

            asyncio.create_task(update_calculation_step())

            calculation_output_data = await gathered

        calculate_step = 1000

        # ■■■■■ get calculation result ■■■■■

        if should_calculate:
            asset_record = previous_asset_record
            for chunk_ouput_data in calculation_output_data:
                chunk_asset_record = chunk_ouput_data["chunk_asset_record"]
                concat_data = [asset_record, chunk_asset_record]
                asset_record = pd.concat(concat_data)
            mask = ~asset_record.index.duplicated()
            asset_record = asset_record[mask]
            asset_record = asset_record.sort_index()

            unrealized_changes = previous_unrealized_changes
            for chunk_ouput_data in calculation_output_data:
                chunk_unrealized_changes = chunk_ouput_data["chunk_unrealized_changes"]
                concat_data = [unrealized_changes, chunk_unrealized_changes]
                unrealized_changes = pd.concat(concat_data)
            mask = ~unrealized_changes.index.duplicated()
            unrealized_changes = unrealized_changes[mask]
            unrealized_changes = unrealized_changes.sort_index()

            scribbles = calculation_output_data[-1]["chunk_scribbles"]
            account_state = calculation_output_data[-1]["chunk_account_state"]
            virtual_state = calculation_output_data[-1]["chunk_virtual_state"]

        else:
            asset_record = previous_asset_record
            unrealized_changes = previous_unrealized_changes
            scribbles = previous_scribbles
            account_state = previous_account_state
            virtual_state = previous_virtual_state

        # ■■■■■ remember and present ■■■■■

        self.raw_asset_record = RWLock(asset_record)
        self.raw_unrealized_changes = RWLock(unrealized_changes)
        self.raw_scribbles = scribbles
        self.raw_account_state = account_state
        self.about_viewing = {
            "year": year,
            "strategy_code_name": strategy_code_name,
            "strategy_version": strategy_version,
        }
        await self.present()

        # ■■■■■ save if properly calculated ■■■■■

        if not only_visible and should_calculate:
            asset_record.to_pickle(asset_record_path)
            unrealized_changes.to_pickle(unrealized_changes_path)
            async with aiofiles.open(scribbles_path, "wb") as file:
                content = pickle.dumps(scribbles)
                await file.write(content)
            async with aiofiles.open(account_state_path, "wb") as file:
                content = pickle.dumps(account_state)
                await file.write(content)
            async with aiofiles.open(virtual_state_path, "wb") as file:
                content = pickle.dumps(virtual_state)
                await file.write(content)

    async def present(self, *args, **kwargs):
        maker_fee = self.presentation_settings["maker_fee"]
        taker_fee = self.presentation_settings["taker_fee"]
        leverage = self.presentation_settings["leverage"]

        async with self.raw_asset_record.read_lock as wrapper:
            asset_record = wrapper.inner.copy()

        async with self.raw_unrealized_changes.read_lock as wrapper:
            unrealized_changes = wrapper.inner.copy()

        scribbles = self.raw_scribbles.copy()
        account_state = self.raw_account_state.copy()

        # ■■■■■ get strategy details ■■■■

        if self.about_viewing is None:
            should_parallelize = False
            chunk_length = 0
        else:
            strategy_index = self.calculation_settings["strategy_index"]
            strategy = solie.window.strategist.strategies[strategy_index]
            should_parallelize = strategy["parallelized_simulation"]
            chunk_length = strategy["chunk_division"]

        # ■■■■■ apply other factors to the asset trace ■■■■

        if should_parallelize:
            division = timedelta(days=chunk_length)
            grouper = pd.Grouper(freq=division, origin="epoch")  # type:ignore
            grouped = asset_record.groupby(grouper)
            chunk_asset_record_list = [r.dropna() for _, r in grouped]
            chunk_count = len(chunk_asset_record_list)

        else:
            chunk_asset_record_list = [asset_record]
            chunk_count = 1

        chunk_asset_changes_list = []
        for turn in range(chunk_count):
            chunk_asset_record = chunk_asset_record_list[turn]
            # leverage
            chunk_result_asset_sr = chunk_asset_record["Result Asset"]
            chunk_asset_shifts = chunk_result_asset_sr.diff()
            if len(chunk_asset_shifts) > 0:
                chunk_asset_shifts.iloc[0] = 0
            lazy_chunk_result_asset = chunk_result_asset_sr.shift(periods=1)
            if len(lazy_chunk_result_asset) > 0:
                lazy_chunk_result_asset.iloc[0] = 1
            chunk_asset_changes_by_leverage = (
                1 + chunk_asset_shifts / lazy_chunk_result_asset * leverage
            )
            # fee
            chunk_fees = chunk_asset_record["Role"].copy()
            chunk_fees[chunk_fees == "maker"] = maker_fee
            chunk_fees[chunk_fees == "taker"] = taker_fee
            chunk_fees = chunk_fees.astype(np.float32)
            chunk_margin_ratios = chunk_asset_record["Margin Ratio"]
            chunk_asset_changes_by_fee = (
                1 - (chunk_fees / 100) * chunk_margin_ratios * leverage
            )
            # altogether
            chunk_asset_changes = (
                chunk_asset_changes_by_leverage * chunk_asset_changes_by_fee
            )
            chunk_asset_changes_list.append(chunk_asset_changes)

        unrealized_changes = unrealized_changes * leverage
        year_asset_changes = pd.concat(chunk_asset_changes_list).sort_index()

        if len(asset_record) > 0:
            start_point = asset_record.index[0]
            year_asset_changes[start_point] = float(1)
            year_asset_changes = year_asset_changes.sort_index()
        asset_record = asset_record.reindex(year_asset_changes.index)
        asset_record["Result Asset"] = year_asset_changes.cumprod()

        presentation_asset_record = asset_record.copy()
        presentation_unrealized_changes = unrealized_changes.copy()
        presentation_scribbles = scribbles.copy()
        presentation_account_state = account_state.copy()

        # ■■■■■ remember ■■■■■

        self.scribbles = presentation_scribbles
        self.account_state = presentation_account_state
        async with self.unrealized_changes.write_lock as wrapper:
            wrapper.replace(presentation_unrealized_changes)
        async with self.asset_record.write_lock as wrapper:
            wrapper.replace(presentation_asset_record)

        # ■■■■■ display ■■■■■

        asyncio.create_task(self.display_lines())
        asyncio.create_task(self.display_range_information())

        if self.about_viewing is None:
            text = "Nothing drawn"
            solie.window.label_19.setText(text)
        else:
            year = self.about_viewing["year"]
            strategy_code_name = self.about_viewing["strategy_code_name"]
            strategy_version = self.about_viewing["strategy_version"]
            text = ""
            text += f"Target year {year}"
            text += "  ⦁  "
            text += f"Strategy code name {strategy_code_name}"
            text += "  ⦁  "
            text += f"Strategy version {strategy_version}"
            solie.window.label_19.setText(text)

    async def display_year_range(self, *args, **kwargs):
        range_start = datetime(
            year=self.calculation_settings["year"],
            month=1,
            day=1,
            tzinfo=timezone.utc,
        )
        range_start = range_start.timestamp()
        range_end = datetime(
            year=self.calculation_settings["year"] + 1,
            month=1,
            day=1,
            tzinfo=timezone.utc,
        )
        range_end = range_end.timestamp()
        widget = solie.window.plot_widget_2
        widget.setXRange(range_start, range_end)

    async def delete_calculation_data(self, *args, **kwargs):
        year = self.calculation_settings["year"]
        strategy_index = self.calculation_settings["strategy_index"]

        strategy = solie.window.strategist.strategies[strategy_index]
        strategy_code_name = strategy["code_name"]
        strategy_version = strategy["version"]

        path_start = f"{self.workerpath}/{strategy_code_name}_{strategy_version}_{year}"
        asset_record_path = path_start + "_asset_record.pickle"
        unrealized_changes_path = path_start + "_unrealized_changes.pickle"
        scribbles_path = path_start + "_scribbles.pickle"
        account_state_path = path_start + "_account_state.pickle"
        virtual_state_path = path_start + "_virtual_state.pickle"

        does_file_exist = False

        if os.path.exists(asset_record_path):
            does_file_exist = True
        if os.path.exists(unrealized_changes_path):
            does_file_exist = True
        if os.path.exists(scribbles_path):
            does_file_exist = True
        if os.path.exists(account_state_path):
            does_file_exist = True
        if os.path.exists(virtual_state_path):
            does_file_exist = True

        if not does_file_exist:
            question = [
                "No calculation data on this combination",
                f"You should calculate first on year {year} with strategy code name"
                f" {strategy_code_name} version {strategy_version}.",
                ["Okay"],
            ]
            await solie.window.ask(question)
            return
        else:
            question = [
                "Are you sure you want to delete calculation data on this combination?",
                "If you do, you should perform the calculation again to see the"
                f" prediction on year {year} with strategy code name"
                f" {strategy_code_name} version {strategy_version}. Calculation data of"
                " other combinations does not get affected.",
                ["Cancel", "Delete"],
            ]
            answer = await solie.window.ask(question)
            if answer in (0, 1):
                return

        try:
            os.remove(asset_record_path)
        except FileNotFoundError:
            pass
        try:
            os.remove(unrealized_changes_path)
        except FileNotFoundError:
            pass
        try:
            os.remove(scribbles_path)
        except FileNotFoundError:
            pass
        try:
            os.remove(account_state_path)
        except FileNotFoundError:
            pass
        try:
            os.remove(virtual_state_path)
        except FileNotFoundError:
            pass

        await self.erase()

    async def draw(self, *args, **kwargs):
        year = self.calculation_settings["year"]
        strategy_index = self.calculation_settings["strategy_index"]

        strategy = solie.window.strategist.strategies[strategy_index]
        strategy_code_name = strategy["code_name"]
        strategy_version = strategy["version"]

        path_start = f"{self.workerpath}/{strategy_code_name}_{strategy_version}_{year}"
        asset_record_path = path_start + "_asset_record.pickle"
        unrealized_changes_path = path_start + "_unrealized_changes.pickle"
        scribbles_path = path_start + "_scribbles.pickle"
        account_state_path = path_start + "_account_state.pickle"

        try:
            async with self.raw_asset_record.write_lock as wrapper:
                new = pd.read_pickle(asset_record_path)
                wrapper.replace(new)
            async with self.raw_unrealized_changes.write_lock as wrapper:
                new = pd.read_pickle(unrealized_changes_path)
                wrapper.replace(new)
            async with aiofiles.open(scribbles_path, "rb") as file:
                content = await file.read()
                self.raw_scribbles = pickle.loads(content)
            async with aiofiles.open(account_state_path, "rb") as file:
                content = await file.read()
                self.raw_account_state = pickle.loads(content)
            self.about_viewing = {
                "year": year,
                "strategy_code_name": strategy_code_name,
                "strategy_version": strategy_version,
            }
            await self.present()
        except FileNotFoundError:
            question = [
                "No calculation data on this combination",
                f"You should calculate first on year {year} with strategy code name"
                f" {strategy_code_name} version {strategy_version}.",
                ["Okay"],
            ]
            await solie.window.ask(question)
            return

    async def match_graph_range(self, *args, **kwargs):
        range_start = solie.window.plot_widget.getAxis("bottom").range[0]
        range_end = solie.window.plot_widget.getAxis("bottom").range[1]
        widget = solie.window.plot_widget_2
        widget.setXRange(range_start, range_end, padding=0)  # type:ignore

    async def stop_calculation(self, *args, **kwargs):
        stop_flag.make("calculate_simulation")

    async def analyze_unrealized_peaks(self, *args, **kwargs):
        async with self.unrealized_changes.read_lock as wrapper:
            sr = wrapper.inner
            peak_indexes, _ = find_peaks(-sr, distance=3600 / 10)  # type:ignore
            peak_sr = sr.iloc[peak_indexes]
        peak_sr = peak_sr.sort_values().iloc[:12]
        if len(peak_sr) < 12:
            question = [
                "Calculation data is either missing or too short",
                "Cannot get the list of meaningful spots with lowest unrealized"
                " profit.",
                ["Okay"],
            ]
            await solie.window.ask(question)
        else:
            text_lines = [
                f"{index} {peak_value:+.2f}%"
                for index, peak_value in peak_sr.iteritems()
            ]
            question = [
                "Spots with lowest unrealized profit",
                "\n".join(text_lines),
                ["Okay"],
            ]
            await solie.window.ask(question)

    async def toggle_combined_draw(self, *args, **kwargs):
        is_checked = args[0]
        if is_checked:
            self.should_draw_all_years = True
        else:
            self.should_draw_all_years = False
        await self.display_lines()



from PySide6 import QtCore, QtWidgets


class DonationGuide(QtWidgets.QWidget):
    def __init__(self, done_event, payload):
        # ■■■■■ the basic ■■■■■

        super().__init__()

        # ■■■■■ full layout ■■■■■

        full_layout = QtWidgets.QHBoxLayout(self)
        cards_layout = QtWidgets.QVBoxLayout()
        full_layout.addLayout(cards_layout)

        # ■■■■■ spacing ■■■■■

        # spacing
        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)

        # ■■■■■ a card ■■■■■

        # card structure
        card = QtWidgets.QGroupBox()
        card.setFixedWidth(720)
        card_layout = QtWidgets.QVBoxLayout(card)
        card_layout.setContentsMargins(80, 40, 80, 40)
        cards_layout.addWidget(card)

        # donation guide text
        text = "😉 If you are benefiting from"
        text += " Solie's features and find it helpful,"
        text += " why not consider supporting the Solie project?"
        text += " Your generous donations contribute to"
        text += " the growth and development of Solie."
        text += "\n\nIf you feel like so, please consider"
        text += " using the BUSD(BSC) wallet address written below."
        label = QtWidgets.QLabel(text)
        label.setWordWrap(True)
        card_layout.addWidget(label)

        # address text
        text = "0xF9A7E35254cc8A9A9C811849CAF672F10fAB7366"
        lineedit_widget = QtWidgets.QLineEdit(text)
        lineedit_widget.setReadOnly(True)
        lineedit_widget.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        card_layout.addWidget(lineedit_widget)

        # ■■■■■ spacing ■■■■■

        # spacing
        spacer = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        cards_layout.addItem(spacer)



import math
from datetime import datetime, timedelta, timezone
from types import CodeType


def choose(**kwargs):
    # ■■■■■ get data ■■■■■

    target_symbols = kwargs["target_symbols"]
    current_moment = kwargs["current_moment"]
    current_candle_data = kwargs["current_candle_data"]
    current_indicators = kwargs["current_indicators"]
    account_state = kwargs["account_state"]
    scribbles = kwargs["scribbles"]
    decision_script: str | CodeType = kwargs["decision_script"]

    # ■■■■■ decision template ■■■■■

    decision = {}
    for symbol in target_symbols:
        decision[symbol] = {}

    # ■■■■■ write decisions ■■■■■

    namespace = {
        "datetime": datetime,
        "timezone": timezone,
        "timedelta": timedelta,
        "math": math,
        "target_symbols": target_symbols,
        "current_moment": current_moment,
        "current_candle_data": current_candle_data,
        "current_indicators": current_indicators,
        "account_state": account_state,
        "scribbles": scribbles,
        "decision": decision,
    }

    exec(decision_script, namespace)

    # ■■■■■ return decision ■■■■■

    blank_symbols = []
    for symbol, symbol_decision in decision.items():
        if len(symbol_decision) == 0:
            blank_symbols.append(symbol)
    for blank_symbol in blank_symbols:
        decision.pop(blank_symbol)

    return decision, scribbles



import json
import os

import aiofiles
import pandas as pd

from solie.recipe import user_settings


async def do():
    datapath = user_settings.get_app_settings()["datapath"]

    # 5.0: data settings
    try:
        filepath = f"{datapath}/basics.json"
        async with aiofiles.open(filepath, "r", encoding="utf8") as file:
            content = await file.read()
            data_settings = json.loads(content)
        os.remove(filepath)
        filepath = f"{datapath}/data_settings.json"
        async with aiofiles.open(filepath, "w", encoding="utf8") as file:
            content = json.dumps(data_settings, indent=4)
            await file.write(content)
    except Exception:
        pass

    # 5.0: symbol column was added to auto order record
    try:
        filepath = f"{datapath}/transactor/auto_order_record.pickle"
        auto_order_record = pd.read_pickle(filepath)
        if "Symbol" not in auto_order_record.columns:
            auto_order_record["Symbol"] = ""
            auto_order_record.to_pickle(filepath)
    except Exception:
        pass

    # 5.0: solie default strategy now has strategy code SLSLDS
    try:
        filepath = f"{datapath}/transactor/automation_settings.json"
        async with aiofiles.open(filepath, "r", encoding="utf8") as file:
            content = await file.read()
            automation_settings = json.loads(content)
        if automation_settings.get("strategy", None) == 2:
            automation_settings.pop("strategy")
            automation_settings["strategy_code"] = "SLSLDS"
        async with aiofiles.open(filepath, "w", encoding="utf8") as file:
            content = json.dumps(automation_settings, indent=4)
            await file.write(content)
    except Exception:
        pass

    # 6.0: use strategy index instead of strategy code
    try:
        filepath = f"{datapath}/transactor/automation_settings.json"
        async with aiofiles.open(filepath, "r", encoding="utf8") as file:
            content = await file.read()
            automation_settings = json.loads(content)
        if "strategy_code" in automation_settings.keys():
            automation_settings.pop("strategy_code")
        if "strategy_index" not in automation_settings.keys():
            automation_settings["strategy_index"] = 0
        async with aiofiles.open(filepath, "w", encoding="utf8") as file:
            content = json.dumps(automation_settings, indent=4)
            await file.write(content)
    except Exception:
        pass

    # 6.3: auto_trade and manual_trade
    try:
        filepath = f"{datapath}/transactor/asset_record.pickle"
        asset_record = pd.read_pickle(filepath)
        asset_record["Cause"] = asset_record["Cause"].replace("trade", "auto_trade")
        asset_record.to_pickle(filepath)
    except Exception:
        pass



import asyncio
import calendar
import itertools
import math
import os
import random
from collections import deque
from datetime import datetime, timedelta, timezone

import numpy as np
import pandas as pd

import solie
from solie.definition.api_requester import ApiRequester
from solie.definition.api_streamer import ApiStreamer
from solie.definition.rw_lock import RWLock
from solie.overlay.donation_guide import DonationGuide
from solie.overlay.download_fill_option import DownloadFillOption
from solie.recipe import (
    check_internet,
    combine_candle_datas,
    download_aggtrade_data,
    fill_holes_with_aggtrades,
    open_browser,
    remember_task_durations,
    simply_format,
    sort_dataframe,
    standardize,
    stop_flag,
    user_settings,
)


class Collector:
    def __init__(self):
        # ■■■■■ for data management ■■■■■

        self.workerpath = user_settings.get_app_settings()["datapath"] + "/collector"
        os.makedirs(self.workerpath, exist_ok=True)

        # ■■■■■ worker secret memory ■■■■■

        self.secret_memory = {
            "price_precisions": {},
            "markets_gone": [],
        }

        # ■■■■■ remember and display ■■■■■

        self.api_requester = ApiRequester()

        self.aggtrade_candle_sizes = {}
        for symbol in user_settings.get_data_settings()["target_symbols"]:
            self.aggtrade_candle_sizes[symbol] = 0

        # candle data
        self.candle_data = RWLock(standardize.candle_data())

        # realtime data chunks
        field_names = itertools.product(
            user_settings.get_data_settings()["target_symbols"],
            ("Best Bid Price", "Best Ask Price", "Mark Price"),
        )
        field_names = [str(field_name) for field_name in field_names]
        dtype = [(field_name, np.float32) for field_name in field_names]
        dtpye = [("index", "datetime64[ns]")] + dtype
        self.realtime_data_chunks = RWLock(
            deque([np.recarray(shape=(0,), dtype=dtpye) for _ in range(2)], maxlen=64)
        )

        # aggregate trades
        field_names = itertools.product(
            user_settings.get_data_settings()["target_symbols"],
            ("Price", "Volume"),
        )
        field_names = [str(field_name) for field_name in field_names]
        dtype = [(field_name, np.float32) for field_name in field_names]
        dtpye = [("index", "datetime64[ns]")] + dtype
        self.aggregate_trades = RWLock(np.recarray(shape=(0,), dtype=dtpye))

        # ■■■■■ repetitive schedules ■■■■■

        solie.window.scheduler.add_job(
            self.display_status_information,
            trigger="cron",
            second="*",
        )
        solie.window.scheduler.add_job(
            self.fill_candle_data_holes,
            trigger="cron",
            second="*/10",
        )
        solie.window.scheduler.add_job(
            self.add_candle_data,
            trigger="cron",
            second="*/10",
        )
        solie.window.scheduler.add_job(
            self.organize_data,
            trigger="cron",
            minute="*",
        )
        solie.window.scheduler.add_job(
            self.get_exchange_information,
            trigger="cron",
            minute="*",
        )
        solie.window.scheduler.add_job(
            self.save_candle_data,
            trigger="cron",
            hour="*",
        )

        # ■■■■■ websocket streamings ■■■■■

        self.api_streamers = {
            "MARK_PRICE": ApiStreamer(
                "wss://fstream.binance.com/ws/!markPrice@arr@1s",
                self.add_mark_price,
            ),
        }
        for symbol in user_settings.get_data_settings()["target_symbols"]:
            api_streamer = ApiStreamer(
                f"wss://fstream.binance.com/ws/{symbol.lower()}@bookTicker",
                self.add_book_tickers,
            )
            self.api_streamers[f"BOOK_TICKER_{symbol}"] = api_streamer
            api_streamer = ApiStreamer(
                f"wss://fstream.binance.com/ws/{symbol.lower()}@aggTrade",
                self.add_aggregate_trades,
            )
            self.api_streamers[f"AGG_TRADE_{symbol}"] = api_streamer

        # ■■■■■ invoked by the internet connection status change  ■■■■■

        check_internet.add_disconnected_functions(self.clear_aggregate_trades)

    async def load(self, *args, **kwargs):
        # candle data
        years = [
            int(simply_format.numeric(filename))
            for filename in os.listdir(self.workerpath)
            if filename.startswith("candle_data_") and filename.endswith(".pickle")
        ]
        async with self.candle_data.write_lock as wrapper:
            inner = wrapper.inner
            divided_datas = [inner]
            for year in years:
                filepath = f"{self.workerpath}/candle_data_{year}.pickle"
                more_df = pd.read_pickle(filepath)
                divided_datas.append(more_df)
            concatenated = pd.concat(divided_datas)
            if not concatenated.index.is_monotonic_increasing:
                concatenated = concatenated.sort_index()
            wrapper.replace(concatenated)
        await asyncio.sleep(0)

    async def organize_data(self, *args, **kwargs):
        start_time = datetime.now(timezone.utc)

        async with self.candle_data.write_lock as wrapper:
            df = wrapper.inner
            original_index = df.index
            unique_index = original_index.drop_duplicates()
            df = df.reindex(unique_index)
            df = df.sort_index()
            df = df.asfreq("10S")
            df = df.astype(np.float32)
            wrapper.replace(df)

        async with self.realtime_data_chunks.write_lock as wrapper:
            inner = wrapper.inner
            inner[-1].sort(order="index")
            if len(inner[-1]) > 2**16:
                new_chunk = inner[-1][0:0].copy()
                inner.append(new_chunk)
                del new_chunk

        async with self.aggregate_trades.write_lock as wrapper:
            inner = wrapper.inner
            inner.sort(order="index")
            last_index = inner[-1]["index"]
            slice_from = last_index - np.timedelta64(60, "s")
            mask = inner["index"] > slice_from
            wrapper.replace(inner[mask].copy())

        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        remember_task_durations.add("collector_organize_data", duration)

    async def save_candle_data(self, *args, **kwargs):
        # ■■■■■ default values ■■■■■

        current_year = datetime.now(timezone.utc).year
        filepath = f"{self.workerpath}/candle_data_{current_year}.pickle"

        async with self.candle_data.read_lock as wrapper:
            df = wrapper.inner
            year_df = df[df.index.year == current_year].copy()  # type:ignore

        # ■■■■■ make a new file ■■■■■

        year_df.to_pickle(filepath + ".new")

        # ■■■■■ safely replace the existing file ■■■■■

        try:
            os.remove(filepath + ".backup")
        except FileNotFoundError:
            pass

        try:
            os.rename(filepath, filepath + ".backup")
        except FileNotFoundError:
            pass

        try:
            os.rename(filepath + ".new", filepath)
        except FileNotFoundError:
            pass

    async def save_all_years_candle_data(self, *args, **kwargs):
        async with self.candle_data.read_lock as wrapper:
            inner = wrapper.inner
            years_sr = inner.index.year.drop_duplicates()  # type:ignore
            years = years_sr.tolist()

        for year in years:
            async with self.candle_data.read_lock as wrapper:
                df = wrapper.inner
                year_df = df[df.index.year == year].copy()  # type:ignore
            filepath = f"{self.workerpath}/candle_data_{year}.pickle"
            year_df.to_pickle(filepath)

    async def get_exchange_information(self, *args, **kwargs):
        if not check_internet.connected():
            return

        payload = {}
        response = await self.api_requester.binance(
            http_method="GET",
            path="/fapi/v1/exchangeInfo",
            payload=payload,
        )
        about_exchange = response

        for about_symbol in about_exchange["symbols"]:
            symbol = about_symbol["symbol"]

            about_filter = {}
            for filter in about_symbol["filters"]:
                if filter["filterType"] == "PRICE_FILTER":
                    about_filter = filter
                    break

            ticksize = float(about_filter["tickSize"])
            price_precision = int(math.log10(1 / ticksize))
            self.secret_memory["price_precisions"][symbol] = price_precision

    async def fill_candle_data_holes(self, *args, **kwargs):
        # ■■■■■ check internet connection ■■■■■

        if not check_internet.connected():
            return

        # ■■■■■ moments ■■■■■

        current_moment = datetime.now(timezone.utc).replace(microsecond=0)
        current_moment = current_moment - timedelta(seconds=current_moment.second % 10)
        split_moment = current_moment - timedelta(days=2)

        # ■■■■■ fill holes ■■■■■

        markets_gone = []
        full_symbols = []
        request_count = 0

        # only the recent part
        async with self.candle_data.read_lock as wrapper:
            df = wrapper.inner
            recent_candle_data = df[df.index >= split_moment].copy()

        target_symbols = user_settings.get_data_settings()["target_symbols"]
        while len(full_symbols) < len(target_symbols) and request_count < 10:
            for symbol in target_symbols:
                if symbol in full_symbols:
                    continue

                recent_candle_data = await solie.event_loop.run_in_executor(
                    solie.process_pool, sort_dataframe.do, recent_candle_data
                )

                from_moment = current_moment - timedelta(hours=24)
                until_moment = current_moment - timedelta(minutes=1)

                inspect_df = recent_candle_data[symbol][from_moment:until_moment]
                base_index = inspect_df.dropna().index
                temp_sr = pd.Series(0, index=base_index)
                written_moments = len(temp_sr)

                if written_moments == (86400 - 60) / 10 + 1:
                    # case when there are no holes
                    full_symbols.append(symbol)
                    continue

                if from_moment not in temp_sr.index:
                    temp_sr[from_moment] = np.nan
                if until_moment not in temp_sr.index:
                    temp_sr[until_moment] = np.nan
                temp_sr = temp_sr.asfreq("10S")
                isnan_sr = temp_sr.isna()
                nan_index = isnan_sr[isnan_sr == 1].index
                moment_to_fill_from: datetime = nan_index[0]  # type:ignore

                # request historical aggtrade data
                aggtrades = {}
                last_fetched_time = moment_to_fill_from
                while last_fetched_time < moment_to_fill_from + timedelta(seconds=10):
                    # intend to fill at least one 10 second candle bar
                    payload = {
                        "symbol": symbol,
                        "startTime": int(last_fetched_time.timestamp() * 1000),
                        "limit": 1000,
                    }
                    response = await self.api_requester.binance(
                        http_method="GET",
                        path="/fapi/v1/aggTrades",
                        payload=payload,
                    )
                    request_count += 1
                    if len(response) == 0:
                        if symbol not in markets_gone:
                            markets_gone.append(symbol)
                        break
                    for aggtrade in response:
                        aggtrade_id = aggtrade["a"]
                        aggtrades[aggtrade_id] = aggtrade
                    last_fetched_id = max(aggtrades.keys())
                    last_fetched_time = datetime.fromtimestamp(
                        aggtrades[last_fetched_id]["T"] / 1000, tz=timezone.utc
                    )

                recent_candle_data = await solie.event_loop.run_in_executor(
                    solie.process_pool,
                    fill_holes_with_aggtrades.do,
                    symbol,
                    recent_candle_data,
                    aggtrades,
                    moment_to_fill_from,
                    last_fetched_time,
                )
        self.secret_memory["markets_gone"] = markets_gone

        # combine
        async with self.candle_data.write_lock as wrapper:
            df = wrapper.inner
            original_candle_data = df[df.index < split_moment]
            # in case the other data is added during the task
            # read the data again
            temp_df = df[df.index >= split_moment]
            recent_candle_data = recent_candle_data.combine_first(temp_df)
            recent_candle_data = recent_candle_data.sort_index()
            candle_data = pd.concat([original_candle_data, recent_candle_data])
            wrapper.replace(candle_data)

    async def display_status_information(self, *args, **kwargs):
        async with self.candle_data.read_lock as wrapper:
            inner = wrapper.inner
            if len(inner) == 0:
                # when the app is executed for the first time
                return

        if len(self.secret_memory["price_precisions"]) == 0:
            # right after the app execution
            return

        # price
        async with self.aggregate_trades.read_lock as wrapper:
            ar = wrapper.inner.copy()
        price_precisions = self.secret_memory["price_precisions"]

        for symbol in user_settings.get_data_settings()["target_symbols"]:
            temp_ar = ar[str((symbol, "Price"))]
            temp_ar = temp_ar[temp_ar != 0]
            if len(temp_ar) > 0:
                price_precision = price_precisions[symbol]
                latest_price = temp_ar[-1]
                text = f"＄{latest_price:.{price_precision}f}"
            else:
                text = "Unavailable"
            solie.window.price_labels[symbol].setText(text)

        # bottom information
        if len(self.secret_memory["markets_gone"]) == 0:
            cumulation_rate = await self.get_candle_data_cumulation_rate()
            async with self.realtime_data_chunks.read_lock as wrapper:
                chunk_count = len(wrapper.inner)
            first_written_time = None
            last_written_time = None
            for turn in range(chunk_count):
                async with self.realtime_data_chunks.read_lock as wrapper:
                    inner = wrapper.inner
                    if len(inner[turn]) > 0:
                        if first_written_time is None:
                            first_record = inner[turn][0]
                            first_written_time = first_record["index"]
                            del first_record
                        last_record = inner[turn][-1]
                        last_written_time = last_record["index"]
                        del last_record
            if first_written_time is not None and last_written_time is not None:
                written_seconds = last_written_time - first_written_time
                written_seconds = written_seconds.astype(np.int64) / 10**9
            else:
                written_seconds = 0
            written_length = timedelta(seconds=written_seconds)
            range_days = written_length.days
            range_hours, remains = divmod(written_length.seconds, 3600)
            range_minutes, remains = divmod(remains, 60)
            written_length_text = f"{range_days}d {range_hours}h {range_minutes}m"

            text = ""
            text += f"24h candle data accumulation rate {cumulation_rate * 100:.2f}%"
            text += "  ⦁  "
            text += f"Realtime data length {written_length_text}"
        else:
            markets_gone = self.secret_memory["markets_gone"]
            if len(markets_gone) == 1:
                text = f"It seems that {markets_gone[0]} market is removed by Binance. You should make a new data folder."
            else:
                text = f"It seems that {', '.join(markets_gone)} markets are removed by Binance. You should make a new data folder."

        solie.window.label_6.setText(text)

    async def get_candle_data_cumulation_rate(self, *args, **kwargs):
        current_moment = datetime.now(timezone.utc).replace(microsecond=0)
        current_moment = current_moment - timedelta(seconds=current_moment.second % 10)
        count_start_moment = current_moment - timedelta(hours=24)
        async with self.candle_data.read_lock as wrapper:
            df = wrapper.inner
            cumulated_moments = len(df[count_start_moment:].dropna())
        needed_moments = 24 * 60 * 60 / 10
        cumulation_rate = min(float(1), (cumulated_moments + 1) / needed_moments)
        return cumulation_rate

    async def open_binance_data_page(self, *args, **kwargs):
        open_browser.do("https://www.binance.com/en/landing/data")

    async def download_fill_candle_data(self, *args, **kwargs):
        # ■■■■■ ask filling type ■■■■■

        answer_container = {"filling_type": None}

        formation = [
            "Choose the range to fill",
            DownloadFillOption,
            True,
            [answer_container],
        ]

        await solie.window.overlay(formation)

        filling_type = answer_container["filling_type"]
        if filling_type is None:
            return

        # ■■■■■ prepare target tuples for downloading ■■■■■

        task_id = stop_flag.make("download_fill_candle_data")

        target_tuples = []
        target_symbols = user_settings.get_data_settings()["target_symbols"]
        if filling_type == 0:
            current_year = datetime.now(timezone.utc).year
            for year in range(2020, current_year):
                for month in range(1, 12 + 1):
                    for symbol in target_symbols:
                        target_tuples.append(
                            (
                                symbol,
                                "monthly",
                                year,
                                month,
                            )
                        )
        if filling_type == 1:
            current_year = datetime.now(timezone.utc).year
            for year in range(2020, current_year):
                for month in range(1, 12 + 1):
                    days_in_month = calendar.monthrange(2019, 1)[1]
                    for day in range(1, days_in_month + 1):
                        for symbol in target_symbols:
                            target_tuples.append(
                                (
                                    symbol,
                                    "daily",
                                    year,
                                    month,
                                    day,
                                )
                            )
        if filling_type == 2:
            current_year = datetime.now(timezone.utc).year
            current_month = datetime.now(timezone.utc).month
            for month in range(1, current_month):
                for symbol in target_symbols:
                    target_tuples.append(
                        (
                            symbol,
                            "monthly",
                            current_year,
                            month,
                        )
                    )
        if filling_type == 3:
            current_year = datetime.now(timezone.utc).year
            current_month = datetime.now(timezone.utc).month
            for month in range(1, current_month):
                days_in_month = calendar.monthrange(2019, 1)[1]
                for day in range(1, days_in_month + 1):
                    for symbol in target_symbols:
                        target_tuples.append(
                            (
                                symbol,
                                "daily",
                                current_year,
                                month,
                                day,
                            )
                        )
        elif filling_type == 4:
            current_year = datetime.now(timezone.utc).year
            current_month = datetime.now(timezone.utc).month
            current_day = datetime.now(timezone.utc).day
            for target_day in range(1, current_day):
                for symbol in target_symbols:
                    target_tuples.append(
                        (
                            symbol,
                            "daily",
                            current_year,
                            current_month,
                            target_day,
                        )
                    )
        elif filling_type == 5:
            now = datetime.now(timezone.utc)
            yesterday = now - timedelta(hours=24)
            day_before_yesterday = yesterday - timedelta(hours=24)
            for symbol in target_symbols:
                target_tuples.append(
                    (
                        symbol,
                        "daily",
                        day_before_yesterday.year,
                        day_before_yesterday.month,
                        day_before_yesterday.day,
                    ),
                )
                target_tuples.append(
                    (
                        symbol,
                        "daily",
                        yesterday.year,
                        yesterday.month,
                        yesterday.day,
                    ),
                )

        total_steps = len(target_tuples)
        done_steps = 0

        # ■■■■■ play the progress bar ■■■■■

        async def play_progress_bar():
            while True:
                if stop_flag.find("download_fill_candle_data", task_id):
                    solie.window.progressBar_3.setValue(0)
                    return
                else:
                    if done_steps == total_steps:
                        progressbar_value = solie.window.progressBar_3.value()
                        if progressbar_value == 1000:
                            await asyncio.sleep(0.1)
                            solie.window.progressBar_3.setValue(0)
                            return
                    before_value = solie.window.progressBar_3.value()
                    if before_value < 1000:
                        remaining = (
                            math.ceil(1000 / total_steps * done_steps) - before_value
                        )
                        new_value = before_value + math.ceil(remaining * 0.2)
                        solie.window.progressBar_3.setValue(new_value)
                    await asyncio.sleep(0.01)

        asyncio.create_task(play_progress_bar())

        # ■■■■■ calculate in parellel ■■■■■

        random.shuffle(target_tuples)
        lanes = solie.process_count
        chunk_size = math.ceil(len(target_tuples) / lanes)
        target_tuple_chunks = []
        for turn in range(lanes):
            new_chunk = target_tuples[turn * chunk_size : (turn + 1) * chunk_size]
            target_tuple_chunks.append(new_chunk)

        async with self.candle_data.read_lock as wrapper:
            inner = wrapper.inner
            combined_df = RWLock(inner.iloc[0:0].copy())

        async def download_fill(target_tuple_chunk):
            nonlocal done_steps
            nonlocal combined_df

            for target_tuple in target_tuple_chunk:
                if stop_flag.find("download_fill_candle_data", task_id):
                    return
                returned = await solie.event_loop.run_in_executor(
                    solie.process_pool,
                    download_aggtrade_data.do,
                    target_tuple,
                )
                if returned is not None:
                    new_df = returned
                    async with combined_df.write_lock as wrapper:
                        inner = wrapper.inner
                        new = await solie.event_loop.run_in_executor(
                            solie.process_pool,
                            combine_candle_datas.do,
                            new_df,
                            inner,
                        )
                        wrapper.replace(new)
                done_steps += 1

        await asyncio.gather(*[download_fill(chunk) for chunk in target_tuple_chunks])

        if stop_flag.find("download_fill_candle_data", task_id):
            return

        # ■■■■■ combine ■■■■■

        async with combined_df.read_lock as wrapper:
            combined_df_inner = wrapper.inner

        async with self.candle_data.write_lock as wrapper:
            inner = wrapper.inner
            df = await solie.event_loop.run_in_executor(
                solie.process_pool,
                combine_candle_datas.do,
                combined_df_inner,
                inner,
            )
            wrapper.replace(df)

        # ■■■■■ save ■■■■■

        await self.save_all_years_candle_data()

        # ■■■■■ add to log ■■■■■

        text = "Filled the candle data with the history data downloaded from Binance"
        solie.logger.info(text)

        # ■■■■■ display to graphs ■■■■■

        asyncio.create_task(solie.window.transactor.display_lines())
        asyncio.create_task(solie.window.simulator.display_lines())

    async def add_book_tickers(self, *args, **kwargs):
        received: dict = kwargs.get("received")  # type:ignore
        start_time = datetime.now(timezone.utc)
        symbol = received["s"]
        best_bid = received["b"]
        best_ask = received["a"]
        event_time = np.datetime64(received["E"] * 10**6, "ns")
        async with self.realtime_data_chunks.write_lock as wrapper:
            inner = wrapper.inner
            original_size = inner[-1].shape[0]
            inner[-1].resize(original_size + 1, refcheck=False)
            inner[-1][-1]["index"] = event_time
            find_key = str((symbol, "Best Bid Price"))
            inner[-1][-1][find_key] = best_bid
            find_key = str((symbol, "Best Ask Price"))
            inner[-1][-1][find_key] = best_ask
        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        remember_task_durations.add("add_book_tickers", duration)

    async def add_mark_price(self, *args, **kwargs):
        received: dict = kwargs.get("received")  # type:ignore
        start_time = datetime.now(timezone.utc)
        target_symbols = user_settings.get_data_settings()["target_symbols"]
        event_time = np.datetime64(received[0]["E"] * 10**6, "ns")
        filtered_data = {}
        for about_mark_price in received:
            symbol = about_mark_price["s"]
            if symbol in target_symbols:
                mark_price = float(about_mark_price["p"])
                filtered_data[symbol] = mark_price
        async with self.realtime_data_chunks.write_lock as wrapper:
            inner = wrapper.inner
            original_size = inner[-1].shape[0]
            inner[-1].resize(original_size + 1, refcheck=False)
            inner[-1][-1]["index"] = event_time
            for symbol, mark_price in filtered_data.items():
                find_key = str((symbol, "Mark Price"))
                inner[-1][-1][find_key] = mark_price
        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        remember_task_durations.add("add_mark_price", duration)

    async def add_aggregate_trades(self, *args, **kwargs):
        received: dict = kwargs.get("received")  # type:ignore
        start_time = datetime.now(timezone.utc)
        symbol = received["s"]
        price = float(received["p"])
        volume = float(received["q"])
        trade_time = np.datetime64(received["T"] * 10**6, "ns")
        async with self.aggregate_trades.write_lock as wrapper:
            inner = wrapper.inner
            original_size = inner.shape[0]
            inner.resize(original_size + 1, refcheck=False)
            inner[-1]["index"] = trade_time
            inner[-1][str((symbol, "Price"))] = price
            inner[-1][str((symbol, "Volume"))] = volume
        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        remember_task_durations.add("add_aggregate_trades", duration)

    async def clear_aggregate_trades(self, *args, **kwargs):
        async with self.aggregate_trades.write_lock as wrapper:
            inner = wrapper.inner
            new = inner[0:0].copy()
            wrapper.replace(new)

    async def add_candle_data(self, *args, **kwargs):
        current_moment = datetime.now(timezone.utc).replace(microsecond=0)
        current_moment = current_moment - timedelta(seconds=current_moment.second % 10)
        before_moment = current_moment - timedelta(seconds=10)

        async with self.aggregate_trades.read_lock as wrapper:
            data_length = len(wrapper.inner)
        if data_length == 0:
            return

        for _ in range(20):
            async with self.aggregate_trades.read_lock as wrapper:
                last_received_index = wrapper.inner[-1]["index"]
            if np.datetime64(current_moment) < last_received_index:
                break
            await asyncio.sleep(0.1)

        async with self.aggregate_trades.read_lock as wrapper:
            aggregate_trades = wrapper.inner.copy()

        first_received_index = aggregate_trades[0]["index"]
        if first_received_index >= np.datetime64(before_moment):
            return

        new_datas = {}

        for symbol in user_settings.get_data_settings()["target_symbols"]:
            block_start_timestamp = before_moment.timestamp()
            block_end_timestamp = current_moment.timestamp()

            index_ar = aggregate_trades["index"].astype(np.int64) / 10**9
            after_start_mask = block_start_timestamp <= index_ar
            before_end_mask = index_ar < block_end_timestamp
            block_ar = aggregate_trades[after_start_mask & before_end_mask]
            block_ar = block_ar[block_ar[str((symbol, "Volume"))] != 0]
            self.aggtrade_candle_sizes[symbol] = block_ar.size

            if len(block_ar) > 0:
                open_price = block_ar[0][str((symbol, "Price"))]
                high_price = block_ar[str((symbol, "Price"))].max()
                low_price = block_ar[str((symbol, "Price"))].min()
                close_price = block_ar[-1][str((symbol, "Price"))]
                sum_volume = block_ar[str((symbol, "Volume"))].sum()
            else:
                async with self.candle_data.read_lock as wrapper:
                    df = wrapper.inner
                    inspect_sr = df.iloc[-60:][(symbol, "Close")].copy()
                inspect_sr = inspect_sr.dropna()
                if len(inspect_sr) == 0:
                    return
                last_price = inspect_sr.tolist()[-1]
                open_price = last_price
                high_price = last_price
                low_price = last_price
                close_price = last_price
                sum_volume = 0

            new_datas[(symbol, "Open")] = open_price
            new_datas[(symbol, "High")] = high_price
            new_datas[(symbol, "Low")] = low_price
            new_datas[(symbol, "Close")] = close_price
            new_datas[(symbol, "Volume")] = sum_volume

        async with self.candle_data.write_lock as wrapper:
            inner = wrapper.inner
            for column_name, new_data_value in new_datas.items():
                inner.loc[before_moment, column_name] = new_data_value
            if not inner.index.is_monotonic_increasing:
                inner = inner.sort_index()

        duration = (datetime.now(timezone.utc) - current_moment).total_seconds()
        remember_task_durations.add("add_candle_data", duration)

    async def stop_filling_candle_data(self, *args, **kwargs):
        stop_flag.make("download_fill_candle_data")

    async def guide_donation(self, *args, **kwargs):
        formation = [
            "Support Solie",
            DonationGuide,
            True,
            None,
        ]
        await solie.window.overlay(formation)



_last_task_id = 0
_flags = {}


def make(task_name: str) -> int:
    global _last_task_id
    new_task_id = _last_task_id + 1
    _flags[task_name] = new_task_id
    _last_task_id = new_task_id
    return new_task_id


def find(task_name: str, current_task_id: int) -> bool:
    if task_name in _flags.keys():
        if _flags[task_name] > current_task_id:
            return True
        else:
            return False
    else:
        return False



import math


def ceil(number: float, decimals: int) -> int:
    number = number * 10**decimals
    number = math.ceil(number)
    number = number / 10**decimals
    number = int(number)
    return number


def floor(number: float, decimals: int) -> int:
    number = number * 10**decimals
    number = math.floor(number)
    number = number / 10**decimals
    number = int(number)
    return number



import itertools
from datetime import datetime, timedelta, timezone
from types import CodeType

import numpy as np
import pandas as pd
import pandas_ta as ta


def do(**kwargs) -> pd.DataFrame:
    # ■■■■■ get data ■■■■■

    target_symbols = kwargs["target_symbols"]
    candle_data = kwargs["candle_data"]
    indicators_script: str | CodeType = kwargs["indicators_script"]

    # ■■■■■ interpolate nans ■■■■■

    candle_data = candle_data.interpolate()

    # ■■■■■ make dummy row to avoid ta error with all nan series ■■■■■

    if len(candle_data) > 0:
        dummy_index = candle_data.index[-1] + timedelta(seconds=1)
    else:
        dummy_index = datetime.fromtimestamp(0, tz=timezone.utc)

    candle_data.loc[dummy_index] = 0

    # ■■■■■ basic values ■■■■■

    blank_columns = itertools.product(
        target_symbols,
        ("Price", "Volume", "Abstract"),
        ("Blank",),
    )
    new_indicators = {}
    base_index = candle_data.index
    for blank_column in blank_columns:
        new_indicators[blank_column] = pd.Series(
            np.nan,
            index=base_index,
            dtype=np.float32,
        )

    # ■■■■■ make individual indicators ■■■■■

    namespace = {
        "ta": ta,
        "pd": pd,
        "np": np,
        "target_symbols": target_symbols,
        "candle_data": candle_data,
        "new_indicators": new_indicators,
    }
    exec(indicators_script, namespace)

    # ■■■■■ concatenate individual indicators into one ■■■■■

    for column_name, new_indicator in new_indicators.items():
        new_indicator.name = column_name

    indicators = pd.concat(new_indicators.values(), axis="columns")
    indicators = indicators.astype(np.float32)

    # ■■■■■ remove dummy row ■■■■■

    indicators = indicators.iloc[:-1]

    return indicators



from PySide6 import QtCore, QtGui, QtWidgets

import solie
from solie.overlay.long_text_view import LongTextView
from solie.recipe import outsource


class LogList(QtWidgets.QListWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.fixed_width_font = QtGui.QFont("Source Code Pro", 9)
        self.setFont(self.fixed_width_font)
        outsource.do(self.itemClicked, self.show_fulltext)

    def addItem(self, summarization: str, log_content: str):  # noqa:N802
        maximum_item_limit = 1024

        new_item = QtWidgets.QListWidgetItem(self)
        new_item.setText(summarization)
        new_item.setData(QtCore.Qt.ItemDataRole.UserRole, log_content)

        super().addItem(new_item)

        index_count = self.count()

        if index_count > maximum_item_limit:
            remove_count = index_count - maximum_item_limit
            for _ in range(remove_count):
                self.takeItem(0)

    async def show_fulltext(self, *args, **kwargs):
        selected_index = self.currentRow()

        selected_item = self.item(selected_index)
        text = selected_item.data(QtCore.Qt.ItemDataRole.UserRole)

        formation = ["This is the full log", LongTextView, True, [text]]
        await solie.window.overlay(formation)
